(* Content-type: application/vnd.wolfram.mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 8.0' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       157,          7]
NotebookDataLength[     12512,        432]
NotebookOptionsPosition[      8786,        321]
NotebookOutlinePosition[     10287,        358]
CellTagsIndexPosition[     10244,        355]
WindowTitle->LieART: Lie Algebras and Representation Theory - Wolfram \
Mathematica
WindowFrame->Normal*)

(* Beginning of Notebook Content *)
Notebook[{
Cell[" ", "GuideColorBar",
 CellMargins->{{Inherited, Inherited}, {-5, 0}}],

Cell[TextData[{
 StyleBox[ButtonBox["Mathematica",
  BaseStyle->{"Link", "LinkTrail"},
  ButtonData->"paclet:guide/Mathematica"],
  FontSlant->"Italic"],
 StyleBox[" > ", "LinkTrailSeparator"]
}], "LinkTrail"],

Cell[CellGroupData[{

Cell["LieART: Lie Algebras and Representation Theory", "GuideTitle",
 CellID->942062912],

Cell["\<\
The primary purpose of the LieART application is the tensor product and \
subalgebra decomposition of irreducible representations (irreps) of Lie \
algebras. These tasks are frequently needed in particle physics, where \
multiparticle fields are assigned to irreps. Tensor products of irreps are \
thus commonly used to find singlets suitable for the Lagrangian. Spontaneous \
symmetry breaking corresponds to the subalgebra decomposition.\
\>", "GuideAbstract",
 CellID->2001916300],

Cell[CellGroupData[{

Cell["Algebras", "GuideFunctionsSubsection",
 CellID->1219780944],

Cell[TextData[{
 Cell[TextData[ButtonBox["Algebra",
  BaseStyle->"Link",
  ButtonData->"paclet:LieART/ref/Algebra"]], "InlineFunctionSans"],
 "\[NonBreakingSpace]",
 StyleBox["\[MediumSpace]\[FilledVerySmallSquare]\[MediumSpace]", 
  "InlineSeparator"],
 " ",
 Cell[TextData[ButtonBox["ProductAlgebra",
  BaseStyle->"Link",
  ButtonData->"paclet:LieART/ref/ProductAlgebra"]], "InlineFunctionSans"],
 "\[NonBreakingSpace]",
 StyleBox["\[MediumSpace]\[FilledVerySmallSquare]\[MediumSpace]", 
  "InlineSeparator"],
 " ",
 Cell[TextData[ButtonBox["CartanMatrix",
  BaseStyle->"Link",
  ButtonData->"paclet:LieART/ref/CartanMatrix"]], "InlineFunctionSans"],
 "\[NonBreakingSpace]",
 StyleBox["\[MediumSpace]\[FilledVerySmallSquare]\[MediumSpace]", 
  "InlineSeparator"],
 " ",
 Cell[TextData[ButtonBox["MetricTensor",
  BaseStyle->"Link",
  ButtonData->"paclet:LieART/ref/MetricTensor"]], "InlineFunctionSans"]
}], "InlineGuideFunctionListing",
 CellID->329721484]
}, Open  ]],

Cell[CellGroupData[{

Cell["Roots", "GuideFunctionsSubsection",
 CellID->1654006133],

Cell[TextData[{
 Cell[TextData[ButtonBox["RootSystem",
  BaseStyle->"Link",
  ButtonData->"paclet:LieART/ref/RootSystem"]], "InlineFunctionSans"],
 " ",
 StyleBox["\[LongDash]", "GuideEmDash"],
 " ",
 "Complete root system of an algebra"
}], "GuideText",
 CellID->1474453966],

Cell[TextData[{
 Cell[TextData[ButtonBox["PositiveRoots",
  BaseStyle->"Link",
  ButtonData->"paclet:LieART/ref/PositiveRoots"]], "InlineFunctionSans"],
 " ",
 StyleBox["\[LongDash]", "GuideEmDash"],
 " ",
 "Only the positive roots of an algebra"
}], "GuideText",
 CellID->595719226],

Cell[TextData[{
 Cell[TextData[ButtonBox["OrthogonalSimpleRoots",
  BaseStyle->"Link",
  ButtonData->"paclet:LieART/ref/OrthogonalSimpleRoots"]], 
  "InlineFunctionSans"],
 " ",
 StyleBox["\[LongDash]", "GuideEmDash"],
 " ",
 "Simple roots of an algebra in orthogonal coordinates"
}], "GuideText",
 CellID->1369365639],

Cell[TextData[Cell[TextData[ButtonBox["HighestRoot",
 BaseStyle->"Link",
 ButtonData->
  "paclet:LieART/ref/HighestRoot"]], "InlineFunctionSans"]], \
"InlineGuideFunctionListing",
 CellID->1900826747]
}, Open  ]],

Cell[CellGroupData[{

Cell["Weights", "GuideFunctionsSubsection",
 CellID->2016205067],

Cell[TextData[{
 Cell[TextData[ButtonBox["WeightSystem",
  BaseStyle->"Link",
  ButtonData->"paclet:LieART/ref/WeightSystem"]], "InlineFunctionSans"],
 " ",
 StyleBox["\[LongDash]", "GuideEmDash"],
 " ",
 "Complete weight system of a representation"
}], "GuideText",
 CellID->1342091672]
}, Open  ]],

Cell[CellGroupData[{

Cell["Representations", "GuideFunctionsSubsection",
 CellID->775384787],

Cell[TextData[{
 Cell[TextData[ButtonBox["Irrep",
  BaseStyle->"Link",
  ButtonData->"paclet:LieART/ref/Irrep"]], "InlineFunctionSans"],
 " ",
 StyleBox["\[LongDash]", "GuideEmDash"],
 " Irreducible representation"
}], "GuideText",
 CellID->361439325],

Cell[TextData[{
 Cell[TextData[ButtonBox["Dim",
  BaseStyle->"Link",
  ButtonData->"paclet:LieART/ref/Dim"]], "InlineFunctionSans"],
 " ",
 StyleBox["\[LongDash]", "GuideEmDash"],
 " ",
 "Compute the dimension of a representation"
}], "GuideText",
 CellID->146551704],

Cell[TextData[{
 Cell[TextData[ButtonBox["YoungTableau",
  BaseStyle->"Link",
  ButtonData->"paclet:LieART/ref/YoungTableau"]], "InlineFunctionSans"],
 " ",
 StyleBox["\[LongDash]", "GuideEmDash"],
 " ",
 "Displays the Young tableau corresponding to an SU(N) representation"
}], "GuideText",
 CellID->447911115],

Cell[TextData[{
 Cell[TextData[ButtonBox["ProductIrrep",
  BaseStyle->"Link",
  ButtonData->"paclet:LieART/ref/ProductIrrep"]], "InlineFunctionSans"],
 "\[NonBreakingSpace]",
 StyleBox["\[MediumSpace]\[FilledVerySmallSquare]\[MediumSpace]", 
  "InlineSeparator"],
 " ",
 Cell[TextData[ButtonBox["DimName",
  BaseStyle->"Link",
  ButtonData->"paclet:LieART/ref/DimName"]], "InlineFunctionSans"],
 "\[NonBreakingSpace]",
 StyleBox["\[MediumSpace]\[FilledVerySmallSquare]\[MediumSpace]", 
  "InlineSeparator"],
 " ",
 Cell[TextData[ButtonBox["Index",
  BaseStyle->"Link",
  ButtonData->"paclet:LieART/ref/Index"]], "InlineFunctionSans"]
}], "InlineGuideFunctionListing",
 CellID->641609725]
}, Open  ]],

Cell[CellGroupData[{

Cell["Decompositions", "GuideFunctionsSubsection",
 CellID->760568900],

Cell[TextData[{
 Cell[TextData[ButtonBox["DecomposeProduct",
  BaseStyle->"Link",
  ButtonData->"paclet:LieART/ref/DecomposeProduct"]], "InlineFunctionSans"],
 " ",
 StyleBox["\[LongDash]", "GuideEmDash"],
 " ",
 "Decompose tensor products of irreps"
}], "GuideText",
 CellID->931894448],

Cell[TextData[{
 Cell[TextData[ButtonBox["DecomposeIrrep",
  BaseStyle->"Link",
  ButtonData->"paclet:LieART/ref/DecomposeIrrep"]], "InlineFunctionSans"],
 " ",
 StyleBox["\[LongDash]", "GuideEmDash"],
 " Decompose irrep to subalgebras"
}], "GuideText",
 CellID->424177561]
}, Open  ]],

Cell[CellGroupData[{

Cell["Displaying Output", "GuideFunctionsSubsection",
 CellID->1103951756],

Cell[TextData[{
 Cell[TextData[ButtonBox["IrrepPlus",
  BaseStyle->"Link",
  ButtonData->"paclet:LieART/ref/IrrepPlus"]], "InlineFunctionSans"],
 " ",
 StyleBox["\[LongDash]", "GuideEmDash"],
 " ",
 "Textbook style sum of a list irreps"
}], "GuideText",
 CellID->1445608321],

Cell[TextData[{
 Cell[TextData[ButtonBox["LaTeXForm",
  BaseStyle->"Link",
  ButtonData->"paclet:LieART/ref/LaTeXForm"]], "InlineFunctionSans"],
 " ",
 StyleBox["\[LongDash]", "GuideEmDash"],
 " ",
 "Formatting suitable for copy&paste into a ",
 Cell[BoxData[
  FormBox[
   StyleBox[
    RowBox[{"L", 
     StyleBox[
      AdjustmentBox["A",
       BoxBaselineShift->-0.4,
       BoxMargins->{{-0.5, -0.3}, {0, 0}}],
      FontSize->Smaller], "T", 
     AdjustmentBox["E",
      BoxBaselineShift->0.5,
      BoxMargins->{{-0.3, 0}, {0, 0}}], "X"}], "TR",
    SingleLetterItalics->False], TraditionalForm]]],
 " document"
}], "GuideText",
 CellID->490235135]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["TUTORIALS", "GuideTutorialsSection",
 CellID->415694126],

Cell[TextData[ButtonBox["Quick Start Tutorial",
 BaseStyle->"Link",
 ButtonData->"paclet:LieART/tutorial/QuickStartTutorial"]], "GuideTutorial",
 CellID->806871991]
}, Open  ]],

Cell[CellGroupData[{

Cell["TABLES",
 Editable->False,
 CellFrame->{{0, 0}, {0, 2}},
 ShowCellBracket->Automatic,
 CellMargins->{{24, 24}, {9, 18}},
 CellElementSpacings->{"ClosedGroupTopMargin"->28},
 CellGroupingRules->{"SectionGrouping", 30},
 CellFrameMargins->4,
 CellFrameColor->RGBColor[
  0.48235300000000003`, 0.5176470000000001, 0.8196080000000001],
 StyleMenuListing->None,
 FontFamily->"Verdana",
 FontSize->9,
 FontTracking->"Wide"],

Cell[TextData[ButtonBox["Representation Properties",
 BaseStyle->"Link",
 ButtonData->"paclet:LieART/Tables/IrrepProperties"]], "GuideMoreAbout",
 CellID->427798698],

Cell[TextData[ButtonBox["Tensor Products",
 BaseStyle->"Link",
 ButtonData->"paclet:LieART/Tables/TensorProducts"]], "GuideMoreAbout",
 CellID->240618869],

Cell[TextData[ButtonBox["Branching Rules",
 BaseStyle->"Link",
 ButtonData->"paclet:LieART/Tables/BranchingRules"]], "GuideMoreAbout",
 CellID->359825597]
}, Open  ]],

Cell[" ", "FooterCell"]
},
Saveable->False,
ScreenStyleEnvironment->"Working",
WindowSize->{725, 750},
WindowMargins->{{0, Automatic}, {Automatic, 0}},
WindowTitle->"LieART: Lie Algebras and Representation Theory - Wolfram \
Mathematica",
TaggingRules->{
 "ModificationHighlight" -> False, 
  "Metadata" -> {
   "built" -> "{2012, 7, 17, 8, 55, 7.668373}", "context" -> "LieART`", 
    "keywords" -> {}, "index" -> True, "label" -> "Lie A R T Package Guide", 
    "language" -> "en", "paclet" -> "LieART Package", "status" -> "None", 
    "summary" -> 
    "The primary purpose of the LieART application is the tensor product and \
subalgebra decomposition of irreducible representations (irreps) of Lie \
algebras. These tasks are frequently needed in particle physics, where \
multiparticle fields are assigned to irreps. Tensor products of irreps are \
thus commonly used to find singlets suitable for the Lagrangian. Spontaneous \
symmetry breaking corresponds to the subalgebra decomposition.", 
    "synonyms" -> {}, "title" -> 
    "LieART: Lie Algebras and Representation Theory", "type" -> "Guide", 
    "uri" -> "LieART/guide/LieART"}, "LinkTrails" -> "", 
  "SearchTextTranslated" -> ""},
FrontEndVersion->"8.0 for Linux x86 (64-bit) (October 10, 2011)",
StyleDefinitions->FrontEnd`FileName[{"Wolfram"}, "Reference.nb", 
  CharacterEncoding -> "UTF-8"]
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{}
*)
(*CellTagsIndex
CellTagsIndex->{}
*)
(*NotebookFileOutline
Notebook[{
Cell[641, 22, 75, 1, 3, "GuideColorBar"],
Cell[719, 25, 209, 6, 19, "LinkTrail"],
Cell[CellGroupData[{
Cell[953, 35, 88, 1, 68, "GuideTitle",
 CellID->942062912],
Cell[1044, 38, 493, 8, 127, "GuideAbstract",
 CellID->2001916300],
Cell[CellGroupData[{
Cell[1562, 50, 65, 1, 30, "GuideFunctionsSubsection",
 CellID->1219780944],
Cell[1630, 53, 959, 26, 21, "InlineGuideFunctionListing",
 CellID->329721484]
}, Open  ]],
Cell[CellGroupData[{
Cell[2626, 84, 62, 1, 30, "GuideFunctionsSubsection",
 CellID->1654006133],
Cell[2691, 87, 275, 9, 22, "GuideText",
 CellID->1474453966],
Cell[2969, 98, 283, 9, 22, "GuideText",
 CellID->595719226],
Cell[3255, 109, 318, 10, 22, "GuideText",
 CellID->1369365639],
Cell[3576, 121, 200, 5, 21, "InlineGuideFunctionListing",
 CellID->1900826747]
}, Open  ]],
Cell[CellGroupData[{
Cell[3813, 131, 64, 1, 30, "GuideFunctionsSubsection",
 CellID->2016205067],
Cell[3880, 134, 287, 9, 22, "GuideText",
 CellID->1342091672]
}, Open  ]],
Cell[CellGroupData[{
Cell[4204, 148, 71, 1, 30, "GuideFunctionsSubsection",
 CellID->775384787],
Cell[4278, 151, 251, 8, 22, "GuideText",
 CellID->361439325],
Cell[4532, 161, 267, 9, 22, "GuideText",
 CellID->146551704],
Cell[4802, 172, 311, 9, 22, "GuideText",
 CellID->447911115],
Cell[5116, 183, 687, 19, 21, "InlineGuideFunctionListing",
 CellID->641609725]
}, Open  ]],
Cell[CellGroupData[{
Cell[5840, 207, 70, 1, 30, "GuideFunctionsSubsection",
 CellID->760568900],
Cell[5913, 210, 287, 9, 22, "GuideText",
 CellID->931894448],
Cell[6203, 221, 273, 8, 22, "GuideText",
 CellID->424177561]
}, Open  ]],
Cell[CellGroupData[{
Cell[6513, 234, 74, 1, 70, "GuideFunctionsSubsection",
 CellID->1103951756],
Cell[6590, 237, 274, 9, 70, "GuideText",
 CellID->1445608321],
Cell[6867, 248, 657, 23, 70, "GuideText",
 CellID->490235135]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[7573, 277, 62, 1, 70, "GuideTutorialsSection",
 CellID->415694126],
Cell[7638, 280, 164, 3, 70, "GuideTutorial",
 CellID->806871991]
}, Open  ]],
Cell[CellGroupData[{
Cell[7839, 288, 423, 13, 70, InheritFromParent,
 CellGroupingRules->{"SectionGrouping", 30}],
Cell[8265, 303, 165, 3, 70, "GuideMoreAbout",
 CellID->427798698],
Cell[8433, 308, 154, 3, 70, "GuideMoreAbout",
 CellID->240618869],
Cell[8590, 313, 154, 3, 70, "GuideMoreAbout",
 CellID->359825597]
}, Open  ]],
Cell[8759, 319, 23, 0, 70, "FooterCell"]
}
]
*)

(* End of internal cache information *)

