(* Content-type: application/vnd.wolfram.mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 8.0' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       157,          7]
NotebookDataLength[     10169,        345]
NotebookOptionsPosition[      7522,        259]
NotebookOutlinePosition[      9070,        304]
CellTagsIndexPosition[      8986,        299]
WindowTitle->ProductAlgebra - Wolfram Mathematica
WindowFrame->Normal*)

(* Beginning of Notebook Content *)
Notebook[{
Cell[" ", "SymbolColorBar",
 CellMargins->{{Inherited, Inherited}, {-5, 0}}],

Cell[TextData[{
 ButtonBox["LieART: Lie Algebras and Representation Theory",
  BaseStyle->{"Link", "LinkTrail"},
  ButtonData->"paclet:LieART/guide/LieART"],
 StyleBox[" > ", "LinkTrailSeparator"]
}], "LinkTrail"],

Cell[BoxData[GridBox[{
   {Cell["LIEART PACLET SYMBOL", "PacletNameCell"], Cell[TextData[{
     Cell[BoxData[
      ActionMenuBox[
       FrameBox["\<\"More About \[RightGuillemet]\"\>",
        StripOnInput->
         False], {"\<\"LieART: Lie Algebras and Representation Theory\"\>":>
       Documentation`HelpLookup["paclet:LieART/guide/LieART"]},
       Appearance->None,
       MenuAppearance->Automatic]],
      LineSpacing->{1.4, 0}],
     "\[ThickSpace]\[ThickSpace]\[ThickSpace]\[ThickSpace]\[ThickSpace]\
\[ThickSpace]",
     Cell[BoxData[
      ActionMenuBox[
       FrameBox["\<\"URL \[RightGuillemet]\"\>",
        StripOnInput->False], {"\<\"LieART/ref/ProductAlgebra\"\>":>
       None, "\<\"Copy Mathematica url\"\>":>
       Module[{DocumentationSearch`Private`nb$}, 
        DocumentationSearch`Private`nb$ = NotebookPut[
           Notebook[{
             Cell["LieART/ref/ProductAlgebra"]}, Visible -> False]]; 
        SelectionMove[DocumentationSearch`Private`nb$, All, Notebook]; 
        FrontEndTokenExecute[DocumentationSearch`Private`nb$, "Copy"]; 
        NotebookClose[DocumentationSearch`Private`nb$]; Null], 
       Delimiter, "\<\"Copy web url\"\>":>
       Module[{DocumentationSearch`Private`nb$}, 
        DocumentationSearch`Private`nb$ = NotebookPut[
           Notebook[{
             Cell[
              BoxData[
               MakeBoxes[
                Hyperlink[
                "http://reference.wolfram.com/mathematica/LieART/ref/\
ProductAlgebra.html"], StandardForm]], "Input", TextClipboardType -> 
              "PlainText"]}, Visible -> False]]; 
        SelectionMove[DocumentationSearch`Private`nb$, All, Notebook]; 
        FrontEndTokenExecute[DocumentationSearch`Private`nb$, "Copy"]; 
        NotebookClose[DocumentationSearch`Private`nb$]; 
        Null], "\<\"Go to web url\"\>":>FrontEndExecute[{
         NotebookLocate[{
           URL[
            StringJoin[
            "http://reference.wolfram.com/mathematica/", 
             "LieART/ref/ProductAlgebra", ".html"]], None}]}]},
       Appearance->None,
       MenuAppearance->Automatic]],
      LineSpacing->{1.4, 0}]
    }], "AnchorBar"]}
  }]], "AnchorBarGrid",
 GridBoxOptions->{GridBoxItemSize->{"Columns" -> {
     Scaled[0.65], {
      Scaled[0.34]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
   "RowsIndexed" -> {}}},
 CellID->1],

Cell[CellGroupData[{

Cell["ProductAlgebra", "ObjectName",
 CellID->1224892054],

Cell[BoxData[GridBox[{
   {"", Cell[TextData[{
     Cell[BoxData[
      RowBox[{"ProductAlgebra", "[", 
       StyleBox["algebras", "TI"], "]"}]], "InlineFormula"],
     " \[LineSeparator]gathers simple Lie ",
     Cell[BoxData[
      StyleBox["algebras", "TI"]], "InlineFormula"],
     " to a product algebra."
    }]]}
  }]], "Usage",
 GridBoxOptions->{
 GridBoxBackground->{
  "Columns" -> {{None}}, "ColumnsIndexed" -> {}, "Rows" -> {{None}}, 
   "RowsIndexed" -> {}}},
 CellID->982511436]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[ButtonBox["EXAMPLES",
 BaseStyle->None,
 Appearance->{Automatic, None},
 Evaluator->None,
 Method->"Preemptive",
 ButtonFunction:>(FrontEndExecute[{
    FrontEnd`SelectionMove[
     FrontEnd`SelectedNotebook[], All, ButtonCell], 
    FrontEndToken["OpenCloseGroup"], 
    FrontEnd`SelectionMove[
     FrontEnd`SelectedNotebook[], After, 
     CellContents]}]& )]], "PrimaryExamplesSection",
 CellTags->"PrimaryExamplesSection",
 CellID->793257672],

Cell[CellGroupData[{

Cell[TextData[{
 "Basic Examples",
 "\[NonBreakingSpace]\[NonBreakingSpace]",
 Cell["(1)", "ExampleCount"]
}], "ExampleSection",
 CellID->1573146131],

Cell[BoxData[
 RowBox[{"Needs", "[", "\"\<LieART`\>\"", "]"}]], "Input",
 CellLabel->"In[1]:=",
 CellID->184741312],

Cell["\<\
Entering the product algebra corresponding to the Standard-Model gauge group:\
\
\>", "ExampleText",
 CellID->1691444968],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ProductAlgebra", "[", 
  RowBox[{"SU3", ",", "SU2", ",", "U1"}], "]"}]], "Input",
 CellLabel->"In[2]:=",
 CellID->435008270],

Cell[BoxData[
 FormBox[
  RowBox[{
   FormBox["\<\"SU(3)\"\>",
    TraditionalForm], "\[Times]", 
   FormBox["\<\"SU(2)\"\>",
    TraditionalForm], "\[Times]", 
   FormBox["\<\"U(1)\"\>",
    TraditionalForm]}], TraditionalForm]], "Output",
 ImageSize->{129, 16},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[2]=",
 CellID->538414704]
}, Open  ]],

Cell["\<\
ProductAlgebra is mostly used to specify the semi-simple or non-semisimple \
subalgebra in subalgebra decompositions:\
\>", "ExampleText",
 CellID->172946751],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DecomposeIrrep", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"Irrep", "[", "SU5", "]"}], "[", "10", "]"}], ",", 
   RowBox[{"ProductAlgebra", " ", "[", 
    RowBox[{"SU3", ",", "SU2", ",", "U1"}], "]"}]}], "]"}]], "Input",
 CellLabel->"In[3]:=",
 CellID->1040201664],

Cell[BoxData[
 FormBox[
  RowBox[{
   RowBox[{"(", 
    FormBox[
     StyleBox["1",
      FontWeight->Bold],
     TraditionalForm], ",", 
    FormBox[
     StyleBox["1",
      FontWeight->Bold],
     TraditionalForm], ")", "\[InvisibleSpace]", 
    FormBox["\<\"(-6)\"\>",
     TraditionalForm]}], "+", 
   RowBox[{"(", 
    FormBox[
     StyleBox[
      OverscriptBox["3", "_"],
      FontWeight->Bold],
     TraditionalForm], ",", 
    FormBox[
     StyleBox["1",
      FontWeight->Bold],
     TraditionalForm], ")", "\[InvisibleSpace]", 
    FormBox["\<\"(4)\"\>",
     TraditionalForm]}], "+", 
   RowBox[{"(", 
    FormBox[
     StyleBox["3",
      FontWeight->Bold],
     TraditionalForm], ",", 
    FormBox[
     StyleBox["2",
      FontWeight->Bold],
     TraditionalForm], ")", "\[InvisibleSpace]", 
    FormBox["\<\"(-1)\"\>",
     TraditionalForm]}]}], TraditionalForm]], "Output",
 ImageSize->{204, 16},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[3]=",
 CellID->369550249]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[ButtonBox["MORE ABOUT",
 BaseStyle->None,
 Appearance->{Automatic, None},
 Evaluator->None,
 Method->"Preemptive",
 ButtonFunction:>(FrontEndExecute[{
    FrontEnd`SelectionMove[
     FrontEnd`SelectedNotebook[], All, ButtonCell], 
    FrontEndToken["OpenCloseGroup"], 
    FrontEnd`SelectionMove[
     FrontEnd`SelectedNotebook[], After, 
     CellContents]}]& )]], "MoreAboutSection",
 CellID->38303248],

Cell[TextData[ButtonBox["LieART: Lie Algebras and Representation Theory",
 BaseStyle->"Link",
 ButtonData->"paclet:LieART/guide/LieART"]], "MoreAbout",
 CellID->1684482064]
}, Open  ]],

Cell[" ", "FooterCell"]
},
Saveable->False,
ScreenStyleEnvironment->"Working",
WindowSize->{725, 750},
WindowMargins->{{0, Automatic}, {Automatic, 0}},
WindowTitle->"ProductAlgebra - Wolfram Mathematica",
TaggingRules->{
 "ModificationHighlight" -> False, 
  "Metadata" -> {
   "built" -> "{2012, 7, 17, 8, 55, 45.112933}", "context" -> "LieART`", 
    "keywords" -> {}, "index" -> True, "label" -> "LieART Paclet Symbol", 
    "language" -> "en", "paclet" -> "LieART", "status" -> "None", "summary" -> 
    "ProductAlgebra[algebras] gathers simple Lie algebras to a product \
algebra.", "synonyms" -> {}, "title" -> "ProductAlgebra", "type" -> "Symbol", 
    "uri" -> "LieART/ref/ProductAlgebra"}, "LinkTrails" -> "", 
  "SearchTextTranslated" -> ""},
CellContext->"Global`",
FrontEndVersion->"8.0 for Linux x86 (64-bit) (October 10, 2011)",
StyleDefinitions->Notebook[{
   Cell[
    StyleData[
    StyleDefinitions -> FrontEnd`FileName[{"Wolfram"}, "Reference.nb"]]], 
   Cell[
    StyleData["Input"], CellContext -> "Global`"], 
   Cell[
    StyleData["Output"], CellContext -> "Global`"]}, Visible -> False, 
  FrontEndVersion -> "8.0 for Linux x86 (64-bit) (October 10, 2011)", 
  StyleDefinitions -> "Default.nb"]
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{
 "PrimaryExamplesSection"->{
  Cell[3870, 112, 461, 13, 45, "PrimaryExamplesSection",
   CellTags->"PrimaryExamplesSection",
   CellID->793257672]}
 }
*)
(*CellTagsIndex
CellTagsIndex->{
 {"PrimaryExamplesSection", 8843, 292}
 }
*)
(*NotebookFileOutline
Notebook[{
Cell[607, 21, 76, 1, 3, "SymbolColorBar"],
Cell[686, 24, 213, 5, 19, "LinkTrail"],
Cell[902, 31, 2353, 54, 45, "AnchorBarGrid",
 CellID->1],
Cell[CellGroupData[{
Cell[3280, 89, 57, 1, 54, "ObjectName",
 CellID->1224892054],
Cell[3340, 92, 493, 15, 85, "Usage",
 CellID->982511436]
}, Open  ]],
Cell[CellGroupData[{
Cell[3870, 112, 461, 13, 45, "PrimaryExamplesSection",
 CellTags->"PrimaryExamplesSection",
 CellID->793257672],
Cell[CellGroupData[{
Cell[4356, 129, 149, 5, 33, "ExampleSection",
 CellID->1573146131],
Cell[4508, 136, 115, 3, 28, "Input",
 CellID->184741312],
Cell[4626, 141, 131, 4, 34, "ExampleText",
 CellID->1691444968],
Cell[CellGroupData[{
Cell[4782, 149, 148, 4, 28, "Input",
 CellID->435008270],
Cell[4933, 155, 371, 13, 37, "Output",
 CellID->538414704]
}, Open  ]],
Cell[5319, 171, 168, 4, 49, "ExampleText",
 CellID->172946751],
Cell[CellGroupData[{
Cell[5512, 179, 287, 8, 28, "Input",
 CellID->1040201664],
Cell[5802, 189, 1023, 41, 37, "Output",
 CellID->369550249]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6886, 237, 419, 12, 45, "MoreAboutSection",
 CellID->38303248],
Cell[7308, 251, 172, 3, 18, "MoreAbout",
 CellID->1684482064]
}, Open  ]],
Cell[7495, 257, 23, 0, 42, "FooterCell"]
}
]
*)

(* End of internal cache information *)

