(* Content-type: application/vnd.wolfram.mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 8.0' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       157,          7]
NotebookDataLength[     68077,       2478]
NotebookOptionsPosition[     57551,       2119]
NotebookOutlinePosition[     59530,       2170]
CellTagsIndexPosition[     59487,       2167]
WindowTitle->LieART - Quick Start Tutorial - Wolfram Mathematica
WindowFrame->Normal*)

(* Beginning of Notebook Content *)
Notebook[{
Cell[" ", "TutorialColorBar",
 CellMargins->{{Inherited, Inherited}, {-5, 0}}],

Cell[TextData[{
 ButtonBox["LieART: Lie Algebras and Representation Theory",
  BaseStyle->{"Link", "LinkTrail"},
  ButtonData->"paclet:LieART/guide/LieART"],
 StyleBox[" > ", "LinkTrailSeparator"]
}], "LinkTrail"],

Cell[BoxData[GridBox[{
   {Cell["LIE A R T PACKAGE TUTORIAL", "PacletNameCell"], Cell[TextData[{
     Cell[BoxData[
      ActionMenuBox[
       FrameBox["\<\"Functions \[RightGuillemet]\"\>",
        StripOnInput->False], {"\<\"DecomposeIrrep\"\>":>
       Documentation`HelpLookup[
       "paclet:LieART/ref/DecomposeIrrep"], "\<\"DecomposeProduct\"\>":>
       Documentation`HelpLookup[
       "paclet:LieART/ref/DecomposeProduct"], "\<\"Irrep\"\>":>
       Documentation`HelpLookup[
       "paclet:LieART/ref/Irrep"], "\<\"ProductIrrep\"\>":>
       Documentation`HelpLookup[
       "paclet:LieART/ref/ProductIrrep"], "\<\"YoungTableau\"\>":>
       Documentation`HelpLookup["paclet:LieART/ref/YoungTableau"]},
       Appearance->None,
       MenuAppearance->Automatic]],
      LineSpacing->{1.4, 0}],
     "\[ThickSpace]\[ThickSpace]\[ThickSpace]\[ThickSpace]\[ThickSpace]\
\[ThickSpace]",
     Cell[BoxData[
      ActionMenuBox[
       FrameBox["\<\"URL \[RightGuillemet]\"\>",
        StripOnInput->False], {"\<\"LieART/tutorial/QuickStartTutorial\"\>":>
       None, "\<\"Copy Mathematica url\"\>":>
       Module[{DocumentationSearch`Private`nb$}, 
        DocumentationSearch`Private`nb$ = NotebookPut[
           Notebook[{
             Cell["LieART/tutorial/QuickStartTutorial"]}, Visible -> False]]; 
        SelectionMove[DocumentationSearch`Private`nb$, All, Notebook]; 
        FrontEndTokenExecute[DocumentationSearch`Private`nb$, "Copy"]; 
        NotebookClose[DocumentationSearch`Private`nb$]; Null], 
       Delimiter, "\<\"Copy web url\"\>":>
       Module[{DocumentationSearch`Private`nb$}, 
        DocumentationSearch`Private`nb$ = NotebookPut[
           Notebook[{
             Cell[
              BoxData[
               MakeBoxes[
                Hyperlink[
                "http://reference.wolfram.com/mathematica/LieART/tutorial/\
QuickStartTutorial.html"], StandardForm]], "Input", TextClipboardType -> 
              "PlainText"]}, Visible -> False]]; 
        SelectionMove[DocumentationSearch`Private`nb$, All, Notebook]; 
        FrontEndTokenExecute[DocumentationSearch`Private`nb$, "Copy"]; 
        NotebookClose[DocumentationSearch`Private`nb$]; 
        Null], "\<\"Go to web url\"\>":>FrontEndExecute[{
         NotebookLocate[{
           URL[
            StringJoin[
            "http://reference.wolfram.com/mathematica/", 
             "LieART/tutorial/QuickStartTutorial", ".html"]], None}]}]},
       Appearance->None,
       MenuAppearance->Automatic]],
      LineSpacing->{1.4, 0}]
    }], "AnchorBar"]}
  }]], "AnchorBarGrid",
 GridBoxOptions->{GridBoxItemSize->{"Columns" -> {
     Scaled[0.65], {
      Scaled[0.34]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
   "RowsIndexed" -> {}}},
 CellID->1],

Cell[CellGroupData[{

Cell["LieART - Quick Start Tutorial", "Title",
 CellID->509267359],

Cell["\<\
The application LieART provides functions to perform the most common tasks \
associated with irreducible representations (irreps) of the classical and \
exceptional Lie algebas.\
\>", "Text",
 CellID->1534169418],

Cell["This loads the package:", "MathCaption",
 CellID->1167244701],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"<<", "LieART`"}]], "Input",
 CellLabel->"In[1]:=",
 CellID->1092165306],

Cell[CellGroupData[{

Cell[BoxData[
 FormBox["\<\"LieART 1.0.0\"\>", TraditionalForm]], "Print",
 CellLabel->"During evaluation of In[1]:=",
 CellID->1715216683],

Cell[BoxData[
 FormBox["\<\"last revised 27 June 12\"\>", TraditionalForm]], "Print",
 CellLabel->"During evaluation of In[1]:=",
 CellID->1183076603]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Entering Irreducible Representations", "Section",
 CellID->1449799695],

Cell[TextData[{
 "Irreducible representations (irreps) are internally decribed by their \
Dynkin label with a combined head of ",
 StyleBox["Irrep", "InlineCode"],
 " and the Lie algebra. "
}], "Text",
 CellID->374349434],

Cell[BoxData[GridBox[{
   {
    RowBox[{
     RowBox[{
      ButtonBox["Irrep",
       BaseStyle->"Link",
       ButtonData->"paclet:LieART/ref/Irrep"], "[", 
      StyleBox["algebraClass", "TI"], "]"}], "[", 
     StyleBox["label", "TI"], "]"}], Cell[TextData[{
     "Irrep descibed by its ",
     StyleBox["algebraClass", "TI"],
     " and Dynkin ",
     StyleBox["label.", "TI"]
    }], "TableText"]}
  }]], "DefinitionBox",
 CellID->2096742444],

Cell["Entering irreps by Dynkin label.", "Caption",
 CellID->1891092685],

Cell[TextData[{
 "The ",
 Cell[BoxData[
  StyleBox["algebraClass", "TI"]], "InlineFormula"],
 " follows the Dynkin classification of simple Lie algebras and can only be \
",
 Cell[BoxData[
  ButtonBox["A",
   BaseStyle->"Link",
   ButtonData->"paclet:LieART/ref/A"]], "InlineFormula"],
 ", ",
 Cell[BoxData[
  ButtonBox["B",
   BaseStyle->"Link",
   ButtonData->"paclet:LieART/ref/B"]], "InlineFormula"],
 ", ",
 Cell[BoxData[
  ButtonBox["C",
   BaseStyle->"Link",
   ButtonData->"paclet:ref/C"]], "InlineFormula"],
 ", ",
 Cell[BoxData[
  ButtonBox["D",
   BaseStyle->"Link",
   ButtonData->"paclet:ref/D"]], "InlineFormula"],
 " for the classical algebras and ",
 Cell[BoxData[
  ButtonBox["E6",
   BaseStyle->"Link",
   ButtonData->"paclet:LieART/ref/E6"]], "InlineFormula"],
 ", ",
 Cell[BoxData[
  ButtonBox["E7",
   BaseStyle->"Link",
   ButtonData->"paclet:LieART/ref/E7"]], "InlineFormula"],
 ", ",
 Cell[BoxData[
  ButtonBox["E8",
   BaseStyle->"Link",
   ButtonData->"paclet:LieART/ref/E8"]], "InlineFormula"],
 ", ",
 Cell[BoxData[
  ButtonBox["F4",
   BaseStyle->"Link",
   ButtonData->"paclet:LieART/ref/F4"]], "InlineFormula"],
 " and ",
 Cell[BoxData[
  ButtonBox["G2",
   BaseStyle->"Link",
   ButtonData->"paclet:LieART/ref/G2"]], "InlineFormula"],
 " for the exceptional algebras. The precise classical algebra is determined \
by the length of the Dynkin label."
}], "Text",
 CellID->144687180],

Cell[TextData[{
 "Entering the ",
 Cell[BoxData[
  StyleBox[
   OverscriptBox["10", "_"],
   FontWeight->"Bold"]], "InlineFormula"],
 " of SU(5) by its Dynkin label and algebra class:"
}], "MathCaption",
 CellID->836781195],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"Irrep", "[", "A", "]"}], "[", 
   RowBox[{"0", ",", "0", ",", "1", ",", "0"}], "]"}], "//", 
  "FullForm"}]], "Input",
 CellLabel->"In[1]:=",
 CellID->2058623809],

Cell[BoxData[
 FormBox[
  TagBox[
   StyleBox[
    RowBox[{
     RowBox[{"Irrep", "[", "A", "]"}], "[", 
     RowBox[{"0", ",", "0", ",", "1", ",", "0"}], "]"}],
    ShowSpecialCharacters->False,
    ShowStringCharacters->True,
    NumberMarks->True],
   FullForm], TraditionalForm]], "Output",
 ImageSize->{115, 16},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[1]//FullForm=",
 CellID->710845293]
}, Open  ]],

Cell[TextData[{
 "In ",
 Cell[BoxData[
  ButtonBox["StandardForm",
   BaseStyle->"Link",
   ButtonData->"paclet:ref/StandardForm"]], "InlineFormula"],
 " the irrep is displayed in the textbook notation of Dynkin labels :"
}], "MathCaption",
 CellID->1221898746],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"Irrep", "[", "A", "]"}], "[", 
   RowBox[{"0", ",", "0", ",", "1", ",", "0"}], "]"}], "//", 
  "StandardForm"}]], "Input",
 CellLabel->"In[2]:=",
 CellID->711449157],

Cell[BoxData["\<\"(0010)\"\>"], "Output",
 ImageSize->{45, 16},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[2]//StandardForm=",
 CellID->320085412]
}, Open  ]],

Cell[TextData[{
 "In ",
 Cell[BoxData[
  ButtonBox["TraditionalForm",
   BaseStyle->"Link",
   ButtonData->"paclet:ref/TraditionalForm"]], "InlineFormula"],
 " (default) the irrep is displayed by its dimensional name:"
}], "MathCaption",
 CellID->681119358],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Irrep", "[", "A", "]"}], "[", 
  RowBox[{"0", ",", "0", ",", "1", ",", "0"}], "]"}]], "Input",
 CellLabel->"In[3]:=",
 CellID->937078051],

Cell[BoxData[
 FormBox[
  StyleBox[
   OverscriptBox["10", "_"],
   FontWeight->Bold], TraditionalForm]], "Output",
 ImageSize->{22, 14},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[3]=",
 CellID->829231914]
}, Open  ]],

Cell[TextData[{
 "The default output format type of LieART is ",
 Cell[BoxData[
  ButtonBox["TraditionalForm",
   BaseStyle->"Link",
   ButtonData->"paclet:ref/TraditionalForm"]], "InlineFormula"],
 ". The associated user setting is overwritten for the notebook LieART is \
loaded in. For ",
 Cell[BoxData[
  ButtonBox["StandardForm",
   BaseStyle->"Link",
   ButtonData->"paclet:ref/StandardForm"]], "InlineFormula"],
 " as output format type please set the global variable ",
 StyleBox["$DefaultOutputForm=StandardForm", "InlineCode"],
 "."
}], "Text",
 CellID->319667636],

Cell[TextData[{
 "As an example for entering an irrep of an exceptional algebra, consider the \
",
 Cell[BoxData[
  StyleBox["27",
   FontWeight->"Bold"]], "InlineFormula"],
 " of ",
 Cell[BoxData[
  SubscriptBox["E", "6"]], "InlineFormula"],
 ":"
}], "MathCaption",
 CellID->1881934180],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Irrep", "[", "E6", "]"}], "[", 
  RowBox[{"1", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0"}], 
  "]"}]], "Input",
 CellLabel->"In[4]:=",
 CellID->464169775],

Cell[BoxData[
 FormBox[
  StyleBox["27",
   FontWeight->Bold], TraditionalForm]], "Output",
 ImageSize->{22, 14},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[4]=",
 CellID->1931618218]
}, Open  ]],

Cell["\<\
Irreps may also be entered by their dimensional name (dimname). The package \
transforms the irrep into its Dynkin label. Since the algebra of an irrep of \
a classical Lie algebra becomes ambiguous with only the dimensional name, it \
has to be specified.\
\>", "Text",
 CellGroupingRules->{"GroupTogetherGrouping", 31},
 CellID->1468912962],

Cell[BoxData[GridBox[{
   {
    RowBox[{
     RowBox[{
      ButtonBox["Irrep",
       BaseStyle->"Link",
       ButtonData->"paclet:LieART/ref/Irrep"], "[", 
      StyleBox["algebra", "TI"], "]"}], "[", 
     StyleBox["dimname", "TI"], "]"}], Cell[TextData[{
     "Irrep entered by its ",
     StyleBox["algebra", "TI"],
     " and ",
     StyleBox["dimname.", "TI"]
    }], "TableText"]}
  }]], "DefinitionBox",
 CellID->792167334],

Cell["Entering irreps by dimensional name.", "Caption",
 CellID->1579371899],

Cell[TextData[{
 "Entering the ",
 Cell[BoxData[
  OverscriptBox["10", "_"]], "InlineFormula",
  FontWeight->"Bold"],
 StyleBox[" ",
  FontWeight->"Bold"],
 "of SU(5) by its dimensional name specifying the algebra by its Dynkin \
classification ",
 Cell[BoxData[
  SubscriptBox["A", "4"]], "InlineFormula"],
 ":"
}], "MathCaption",
 CellID->109632626],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"Irrep", "[", "A4", "]"}], "[", 
   RowBox[{"Bar", "[", "10", "]"}], "]"}], "//", "InputForm"}]], "Input",
 CellLabel->"In[5]:=",
 CellID->333587606],

Cell["Irrep[A][0, 0, 1, 0]", "Output",
 ImageSize->{114, 14},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[5]//InputForm=",
 CellID->1988021587]
}, Open  ]],

Cell["The traditional name of the algebra SU(5) may also be used:", \
"MathCaption",
 CellID->1511260992],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"Irrep", "[", "SU5", "]"}], "[", 
   RowBox[{"Bar", "[", "10", "]"}], "]"}], "//", "InputForm"}]], "Input",
 CellLabel->"In[6]:=",
 CellID->1425383264],

Cell["Irrep[A][0, 0, 1, 0]", "Output",
 ImageSize->{114, 14},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[6]//InputForm=",
 CellID->622885791]
}, Open  ]],

Cell[TextData[{
 "Irreps of product algebras like SU(3)\[CircleTimes]SU(2)\[CircleTimes]U(1) \
are specified by ",
 Cell[BoxData[
  ButtonBox["ProductIrrep",
   BaseStyle->"Link",
   ButtonData->"paclet:LieART/ref/ProductIrrep"]], "InlineFormula"],
 " with the individual irreps of simple Lie algebras as arguments."
}], "Text",
 CellID->76966508],

Cell[BoxData[GridBox[{
   {
    RowBox[{
     ButtonBox[
      ButtonBox["ProductIrrep",
       BaseStyle->"Link",
       ButtonData->"paclet:LieART/ref/ProductIrrep"],
      BaseStyle->"Link",
      ButtonData->"paclet:LieART/ref/ProductIrrep"], "[", 
     StyleBox["irreps", "TI"], "]"}], Cell[TextData[{
     "Head of product ",
     StyleBox["irreps", "TI"],
     ", gathering irreps of simple Lie algebras."
    }], "TableText"]}
  }]], "DefinitionBox",
 CellID->616137037],

Cell["Product irreps.", "Caption",
 CellID->722483301],

Cell[TextData[{
 "The product irrep ",
 Cell[BoxData[
  RowBox[{"(", 
   RowBox[{
    StyleBox["3",
     FontWeight->"Bold"], ",", 
    StyleBox["3",
     FontWeight->"Bold"]}], ")"}]], "InlineFormula"],
 " of SU(3)\[CircleTimes]SU(3):"
}], "MathCaption",
 CellID->1501046753],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ProductIrrep", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"Irrep", "[", "SU3", "]"}], "[", "3", "]"}], ",", 
   RowBox[{
    RowBox[{"Irrep", "[", "SU3", "]"}], "[", 
    RowBox[{"Bar", "[", "3", "]"}], "]"}]}], "]"}]], "Input",
 CellLabel->"In[7]:=",
 CellID->1686390499],

Cell[BoxData[
 FormBox[
  RowBox[{"(", 
   FormBox[
    StyleBox["3",
     FontWeight->Bold],
    TraditionalForm], ",", 
   FormBox[
    StyleBox[
     OverscriptBox["3", "_"],
     FontWeight->Bold],
    TraditionalForm], ")"}], TraditionalForm]], "Output",
 ImageSize->{39, 16},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[7]=",
 CellID->1102757659]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "InputForm"}]], "Input",
 CellLabel->"In[8]:=",
 CellID->680021955],

Cell["ProductIrrep[Irrep[A][1, 0], Irrep[A][0, 1]]", "Output",
 ImageSize->{260, 14},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[8]//InputForm=",
 CellID->1927209351]
}, Open  ]],

Cell[CellGroupData[{

Cell["ProductIrrep[Irrep[A][1, 0], Irrep[A][0, 1]]", "Input",
 CellLabel->"In[9]:=",
 CellID->1308426208],

Cell[BoxData[
 FormBox[
  RowBox[{"(", 
   FormBox[
    StyleBox["3",
     FontWeight->Bold],
    TraditionalForm], ",", 
   FormBox[
    StyleBox[
     OverscriptBox["3", "_"],
     FontWeight->Bold],
    TraditionalForm], ")"}], TraditionalForm]], "Output",
 ImageSize->{39, 16},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[9]=",
 CellID->739296930]
}, Open  ]],

Cell["\<\
Take for example the left-handed quark doublet in the Standard-Model gauge \
group SU(3)\[CircleTimes]SU(2)\[CircleTimes]U(1) (The U(1) charge is not \
typeset in bold face):\
\>", "MathCaption",
 CellID->1021666716],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ProductIrrep", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"Irrep", "[", "SU3", "]"}], "[", "3", "]"}], ",", 
   RowBox[{
    RowBox[{"Irrep", "[", "SU2", "]"}], "[", "2", "]"}], ",", 
   RowBox[{
    RowBox[{"Irrep", "[", "U1", "]"}], "[", 
    RowBox[{"1", "/", "3"}], "]"}]}], "]"}]], "Input",
 CellLabel->"In[10]:=",
 CellID->1109740049],

Cell[BoxData[
 FormBox[
  RowBox[{"(", 
   FormBox[
    StyleBox["3",
     FontWeight->Bold],
    TraditionalForm], ",", 
   FormBox[
    StyleBox["2",
     FontWeight->Bold],
    TraditionalForm], ")", "\[InvisibleSpace]", 
   FormBox["\<\"(1/3)\"\>",
    TraditionalForm]}], TraditionalForm]], "Output",
 ImageSize->{69, 16},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[10]=",
 CellID->1746525800]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "InputForm"}]], "Input",
 CellLabel->"In[11]:=",
 CellID->1050432190],

Cell["\<\
ProductIrrep[Irrep[A][1, 0], Irrep[A][1], 
 Irrep[U][1/3]]\
\>", "Output",
 ImageSize->{247, 29},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[11]//InputForm=",
 CellID->1160495567]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Decomposing Tensor Products ", "Section",
 CellGroupingRules->{"SectionGrouping", Inherited},
 CellID->354599395],

Cell[BoxData[GridBox[{
   {
    RowBox[{
     ButtonBox["DecomposeProduct",
      BaseStyle->"Link",
      ButtonData->"paclet:LieART/ref/DecomposeProduct"], "[", 
     StyleBox["irreps", "TI"], "]"}], Cell[TextData[{
     "Decomposes the tensor product of several ",
     StyleBox["irreps.", "TI"]
    }], "TableText"]}
  }]], "DefinitionBox",
 CellID->1309055104],

Cell["Tensor product decomposition.", "Caption",
 CellID->1795200765],

Cell[TextData[{
 "Decompose the tensor product ",
 Cell[BoxData[
  RowBox[{
   StyleBox["3",
    FontWeight->"Bold"], "\[CircleTimes]", 
   StyleBox[
    OverscriptBox["3", "_"],
    FontWeight->"Bold"]}]], "InlineFormula"],
 " of SU(3):"
}], "MathCaption",
 CellID->1189993846],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DecomposeProduct", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"Irrep", "[", "SU3", "]"}], "[", "3", "]"}], ",", 
   RowBox[{
    RowBox[{"Irrep", "[", "SU3", "]"}], "[", 
    RowBox[{"Bar", "[", "3", "]"}], "]"}]}], "]"}]], "Input",
 CellLabel->"In[12]:=",
 CellID->1332771530],

Cell[BoxData[
 FormBox[
  RowBox[{
   FormBox[
    StyleBox["1",
     FontWeight->Bold],
    TraditionalForm], "+", 
   FormBox[
    StyleBox["8",
     FontWeight->Bold],
    TraditionalForm]}], TraditionalForm]], "Output",
 ImageSize->{35, 14},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[12]=",
 CellID->1163229237]
}, Open  ]],

Cell[TextData[{
 "Decompose the tensor product ",
 Cell[BoxData[
  RowBox[{
   StyleBox["27",
    FontWeight->"Bold"], "\[CircleTimes]", 
   StyleBox[
    OverscriptBox["27", "_"],
    FontWeight->"Bold"]}]], "InlineFormula"],
 " of ",
 Cell[BoxData[
  SubscriptBox["E", "6"]], "InlineFormula"],
 ":"
}], "MathCaption",
 CellID->133991959],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DecomposeProduct", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"Irrep", "[", "E6", "]"}], "[", "27", "]"}], ",", " ", 
   RowBox[{
    RowBox[{"Irrep", "[", "E6", "]"}], "[", 
    RowBox[{"Bar", "[", "27", "]"}], "]"}]}], "]"}]], "Input",
 CellLabel->"In[13]:=",
 CellID->101963549],

Cell[BoxData[
 FormBox[
  RowBox[{
   FormBox[
    StyleBox["1",
     FontWeight->Bold],
    TraditionalForm], "+", 
   FormBox[
    StyleBox["78",
     FontWeight->Bold],
    TraditionalForm], "+", 
   FormBox[
    StyleBox["650",
     FontWeight->Bold],
    TraditionalForm]}], TraditionalForm]], "Output",
 ImageSize->{84, 14},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[13]=",
 CellID->600695207]
}, Open  ]],

Cell[TextData[{
 "Decompose the tensor product ",
 Cell[BoxData[
  RowBox[{
   StyleBox["3",
    FontWeight->"Bold"], "\[CircleTimes]", 
   StyleBox["3",
    FontWeight->"Bold"], "\[CircleTimes]", 
   StyleBox["3",
    FontWeight->"Bold"]}]], "InlineFormula"],
 " of SU(3):"
}], "MathCaption",
 CellID->1984894909],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DecomposeProduct", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"Irrep", "[", "SU3", "]"}], "[", "3", "]"}], ",", 
   RowBox[{
    RowBox[{"Irrep", "[", "SU3", "]"}], "[", "3", "]"}], ",", 
   RowBox[{
    RowBox[{"Irrep", "[", "SU3", "]"}], "[", "3", "]"}]}], "]"}]], "Input",
 CellLabel->"In[14]:=",
 CellID->1356021807],

Cell[BoxData[
 FormBox[
  RowBox[{
   FormBox[
    StyleBox["1",
     FontWeight->Bold],
    TraditionalForm], "+", 
   RowBox[{"2", "(", 
    FormBox[
     StyleBox["8",
      FontWeight->Bold],
     TraditionalForm], ")"}], "+", 
   FormBox[
    StyleBox["10",
     FontWeight->Bold],
    TraditionalForm]}], TraditionalForm]], "Output",
 ImageSize->{86, 16},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[14]=",
 CellID->502898059]
}, Open  ]],

Cell[TextData[{
 "Decompose the tensor product ",
 Cell[BoxData[
  RowBox[{
   StyleBox["8",
    FontWeight->"Bold"], "\[CircleTimes]", 
   StyleBox["8",
    FontWeight->"Bold"]}]], "InlineFormula"],
 " of SU(3):"
}], "MathCaption",
 CellID->1005626329],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DecomposeProduct", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"Irrep", "[", "SU3", "]"}], "[", "8", "]"}], ",", 
   RowBox[{
    RowBox[{"Irrep", "[", "SU3", "]"}], "[", "8", "]"}]}], "]"}]], "Input",
 CellLabel->"In[15]:=",
 CellID->1887184167],

Cell[BoxData[
 FormBox[
  RowBox[{
   FormBox[
    StyleBox["1",
     FontWeight->Bold],
    TraditionalForm], "+", 
   RowBox[{"2", "(", 
    FormBox[
     StyleBox["8",
      FontWeight->Bold],
     TraditionalForm], ")"}], "+", 
   FormBox[
    StyleBox["10",
     FontWeight->Bold],
    TraditionalForm], "+", 
   FormBox[
    StyleBox[
     OverscriptBox["10", "_"],
     FontWeight->Bold],
    TraditionalForm], "+", 
   FormBox[
    StyleBox["27",
     FontWeight->Bold],
    TraditionalForm]}], TraditionalForm]], "Output",
 ImageSize->{148, 16},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[15]=",
 CellID->1871092227]
}, Open  ]],

Cell[TextData[{
 "Internally a sum of irreps is represented by ",
 Cell[BoxData[
  ButtonBox["IrrepPlus",
   BaseStyle->"Link",
   ButtonData->"paclet:LieART/ref/IrrepPlus"]], "InlineFormula"],
 " and ",
 Cell[BoxData[
  ButtonBox["IrrepTimes",
   BaseStyle->"Link",
   ButtonData->"paclet:LieART/ref/IrrepTimes"]], "InlineFormula"],
 ", an analog of the built-in functions ",
 Cell[BoxData[
  ButtonBox["Plus",
   BaseStyle->"Link",
   ButtonData->"paclet:ref/Plus"]], "InlineFormula"],
 " and ",
 Cell[BoxData[
  ButtonBox["Times",
   BaseStyle->"Link",
   ButtonData->"paclet:ref/Times"]], "InlineFormula"],
 ":"
}], "MathCaption",
 CellID->683723966],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "InputForm"}]], "Input",
 CellLabel->"In[16]:=",
 CellID->997413459],

Cell["\<\
IrrepPlus[Irrep[A][0, 0], IrrepTimes[2, 
  Irrep[A][1, 1]], Irrep[A][3, 0], Irrep[A][0, 3], 
 Irrep[A][2, 2]]\
\>", "Output",
 ImageSize->{283, 44},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[16]//InputForm=",
 CellID->107401848]
}, Open  ]],

Cell[TextData[{
 "Results can be transformed into a list of irreps with ",
 Cell[BoxData[
  ButtonBox["IrrepList",
   BaseStyle->"Link",
   ButtonData->"paclet:LieART/ref/IrrepList"]], "InlineFormula"],
 ", suitable for further processing with functions like ",
 Cell[BoxData[
  ButtonBox["Select",
   BaseStyle->"Link",
   ButtonData->"paclet:ref/Select"]], "InlineFormula"],
 " or ",
 Cell[BoxData[
  ButtonBox["Cases",
   BaseStyle->"Link",
   ButtonData->"paclet:ref/Cases"]], "InlineFormula"],
 ":"
}], "MathCaption",
 CellID->1581190875],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "IrrepList"}]], "Input",
 CellLabel->"In[17]:=",
 CellID->863826808],

Cell[BoxData[
 FormBox[
  RowBox[{"{", 
   RowBox[{
    FormBox[
     StyleBox["1",
      FontWeight->Bold],
     TraditionalForm], ",", 
    FormBox[
     StyleBox["8",
      FontWeight->Bold],
     TraditionalForm], ",", 
    FormBox[
     StyleBox["8",
      FontWeight->Bold],
     TraditionalForm], ",", 
    FormBox[
     StyleBox["10",
      FontWeight->Bold],
     TraditionalForm], ",", 
    FormBox[
     StyleBox[
      OverscriptBox["10", "_"],
      FontWeight->Bold],
     TraditionalForm], ",", 
    FormBox[
     StyleBox["27",
      FontWeight->Bold],
     TraditionalForm]}], "}"}], TraditionalForm]], "Output",
 ImageSize->{135, 18},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[17]=",
 CellID->1351507635]
}, Open  ]],

Cell["Decompose tensor products of product irreps:", "MathCaption",
 CellID->328377256],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DecomposeProduct", "[", 
  RowBox[{
   RowBox[{"ProductIrrep", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"Irrep", "[", "SU3", "]"}], "[", "3", "]"}], ",", 
     RowBox[{
      RowBox[{"Irrep", "[", "SU3", "]"}], "[", 
      RowBox[{"Bar", "[", "3", "]"}], "]"}], ",", 
     RowBox[{
      RowBox[{"Irrep", "[", "SU3", "]"}], "[", "1", "]"}]}], "]"}], ",", 
   RowBox[{"ProductIrrep", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"Irrep", "[", "SU3", "]"}], "[", 
      RowBox[{"Bar", "[", "3", "]"}], "]"}], ",", 
     RowBox[{
      RowBox[{"Irrep", "[", "SU3", "]"}], "[", "3", "]"}], ",", 
     RowBox[{
      RowBox[{"Irrep", "[", "SU3", "]"}], "[", "1", "]"}]}], "]"}]}], 
  "]"}]], "Input",
 CellLabel->"In[18]:=",
 CellID->451431300],

Cell[BoxData[
 FormBox[
  RowBox[{
   RowBox[{"(", 
    FormBox[
     StyleBox["1",
      FontWeight->Bold],
     TraditionalForm], ",", 
    FormBox[
     StyleBox["1",
      FontWeight->Bold],
     TraditionalForm], ",", 
    FormBox[
     StyleBox["1",
      FontWeight->Bold],
     TraditionalForm], ")"}], "+", 
   RowBox[{"(", 
    FormBox[
     StyleBox["8",
      FontWeight->Bold],
     TraditionalForm], ",", 
    FormBox[
     StyleBox["1",
      FontWeight->Bold],
     TraditionalForm], ",", 
    FormBox[
     StyleBox["1",
      FontWeight->Bold],
     TraditionalForm], ")"}], "+", 
   RowBox[{"(", 
    FormBox[
     StyleBox["1",
      FontWeight->Bold],
     TraditionalForm], ",", 
    FormBox[
     StyleBox["8",
      FontWeight->Bold],
     TraditionalForm], ",", 
    FormBox[
     StyleBox["1",
      FontWeight->Bold],
     TraditionalForm], ")"}], "+", 
   RowBox[{"(", 
    FormBox[
     StyleBox["8",
      FontWeight->Bold],
     TraditionalForm], ",", 
    FormBox[
     StyleBox["8",
      FontWeight->Bold],
     TraditionalForm], ",", 
    FormBox[
     StyleBox["1",
      FontWeight->Bold],
     TraditionalForm], ")"}]}], TraditionalForm]], "Output",
 ImageSize->{251, 16},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[18]=",
 CellID->814463409]
}, Open  ]],

Cell[TextData[{
 "Decompose the tensor products ",
 Cell[BoxData[
  RowBox[{
   RowBox[{"(", 
    RowBox[{
     StyleBox["3",
      FontWeight->"Bold"], ",", 
     StyleBox["2",
      FontWeight->"Bold"]}], ")"}], "\[CircleTimes]", 
   RowBox[{"(", 
    RowBox[{
     StyleBox[
      OverscriptBox["3", "_"],
      FontWeight->"Bold"], ",", 
     StyleBox["1",
      FontWeight->"Bold"]}], ")"}]}]], "InlineFormula"],
 " of SU(3)\[CircleTimes]SU(2):"
}], "MathCaption",
 CellID->1480557077],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DecomposeProduct", "[", 
  RowBox[{
   RowBox[{"ProductIrrep", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"Irrep", "[", "SU3", "]"}], "[", "3", "]"}], ",", 
     RowBox[{
      RowBox[{"Irrep", "[", "SU2", "]"}], "[", "2", "]"}]}], "]"}], ",", 
   RowBox[{"ProductIrrep", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"Irrep", "[", "SU3", "]"}], "[", 
      RowBox[{"Bar", "[", "3", "]"}], "]"}], ",", 
     RowBox[{
      RowBox[{"Irrep", "[", "SU2", "]"}], "[", "1", "]"}]}], "]"}]}], 
  "]"}]], "Input",
 CellLabel->"In[19]:=",
 CellID->343055154],

Cell[BoxData[
 FormBox[
  RowBox[{
   RowBox[{"(", 
    FormBox[
     StyleBox["1",
      FontWeight->Bold],
     TraditionalForm], ",", 
    FormBox[
     StyleBox["2",
      FontWeight->Bold],
     TraditionalForm], ")"}], "+", 
   RowBox[{"(", 
    FormBox[
     StyleBox["8",
      FontWeight->Bold],
     TraditionalForm], ",", 
    FormBox[
     StyleBox["2",
      FontWeight->Bold],
     TraditionalForm], ")"}]}], TraditionalForm]], "Output",
 ImageSize->{87, 16},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[19]=",
 CellID->415443930]
}, Open  ]],

Cell[TextData[{
 "Decompose the tensor product ",
 Cell[BoxData[
  RowBox[{
   StyleBox["4",
    FontWeight->"Bold"], "\[CircleTimes]", 
   StyleBox["4",
    FontWeight->"Bold"], "\[CircleTimes]", 
   StyleBox["6",
    FontWeight->"Bold"], "\[CircleTimes]", 
   StyleBox["15",
    FontWeight->"Bold"]}]], "InlineFormula"],
 " of SU(4):"
}], "MathCaption",
 CellID->846894030],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DecomposeProduct", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"Irrep", "[", "SU4", "]"}], "[", "4", "]"}], ",", " ", 
   RowBox[{
    RowBox[{"Irrep", "[", "SU4", "]"}], "[", "4", "]"}], ",", " ", 
   RowBox[{
    RowBox[{"Irrep", "[", "SU4", "]"}], "[", "6", "]"}], ",", " ", 
   RowBox[{
    RowBox[{"Irrep", "[", "SU4", "]"}], "[", "15", "]"}]}], "]"}]], "Input",
 CellLabel->"In[20]:=",
 CellID->1274938883],

Cell[BoxData[
 FormBox[
  RowBox[{
   RowBox[{"2", "(", 
    FormBox[
     StyleBox["1",
      FontWeight->Bold],
     TraditionalForm], ")"}], "+", 
   RowBox[{"7", "(", 
    FormBox[
     StyleBox["15",
      FontWeight->Bold],
     TraditionalForm], ")"}], "+", 
   RowBox[{"4", "(", 
    FormBox[
     SuperscriptBox[
      StyleBox["20",
       FontWeight->Bold], "\[Prime]"],
     TraditionalForm], ")"}], "+", 
   FormBox[
    StyleBox["35",
     FontWeight->Bold],
    TraditionalForm], "+", 
   RowBox[{"5", "(", 
    FormBox[
     StyleBox["45",
      FontWeight->Bold],
     TraditionalForm], ")"}], "+", 
   RowBox[{"3", "(", 
    FormBox[
     StyleBox[
      OverscriptBox["45", "_"],
      FontWeight->Bold],
     TraditionalForm], ")"}], "+", 
   RowBox[{"3", "(", 
    FormBox[
     StyleBox["84",
      FontWeight->Bold],
     TraditionalForm], ")"}], "+", 
   RowBox[{"2", "(", 
    FormBox[
     StyleBox["175",
      FontWeight->Bold],
     TraditionalForm], ")"}], "+", 
   FormBox[
    StyleBox["256",
     FontWeight->Bold],
    TraditionalForm]}], TraditionalForm]], "Output",
 ImageSize->{423, 18},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[20]=",
 CellID->1329380349]
}, Open  ]],

Cell[TextData[{
 "Decompose the tensor product ",
 Cell[BoxData[
  RowBox[{
   StyleBox[
    OverscriptBox["10", "_"],
    FontWeight->"Bold"], "\[CircleTimes]", 
   StyleBox["24",
    FontWeight->"Bold"], "\[CircleTimes]", 
   StyleBox["45",
    FontWeight->"Bold"]}]], "InlineFormula"],
 " of SU(5):"
}], "MathCaption",
 CellID->163980459],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DecomposeProduct", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"Irrep", "[", "SU5", "]"}], "[", 
    RowBox[{"Bar", "[", "10", "]"}], "]"}], ",", 
   RowBox[{
    RowBox[{"Irrep", "[", "SU5", "]"}], "[", "24", "]"}], ",", 
   RowBox[{
    RowBox[{"Irrep", "[", "SU5", "]"}], "[", "45", "]"}]}], "]"}]], "Input",
 CellLabel->"In[21]:=",
 CellID->62108204],

Cell[BoxData[
 FormBox[
  RowBox[{
   RowBox[{"3", "(", 
    FormBox[
     StyleBox[
      OverscriptBox["5", "_"],
      FontWeight->Bold],
     TraditionalForm], ")"}], "+", 
   RowBox[{"6", "(", 
    FormBox[
     StyleBox[
      OverscriptBox["45", "_"],
      FontWeight->Bold],
     TraditionalForm], ")"}], "+", 
   RowBox[{"3", "(", 
    FormBox[
     StyleBox[
      OverscriptBox["50", "_"],
      FontWeight->Bold],
     TraditionalForm], ")"}], "+", 
   RowBox[{"5", "(", 
    FormBox[
     StyleBox[
      OverscriptBox["70", "_"],
      FontWeight->Bold],
     TraditionalForm], ")"}], "+", 
   RowBox[{"2", "(", 
    FormBox[
     StyleBox[
      OverscriptBox["105", "_"],
      FontWeight->Bold],
     TraditionalForm], ")"}], "+", 
   FormBox[
    SuperscriptBox[
     StyleBox[
      OverscriptBox["175", "_"],
      FontWeight->Bold], "\[Prime]\[Prime]"],
    TraditionalForm], "+", 
   RowBox[{"6", "(", 
    FormBox[
     StyleBox[
      OverscriptBox["280", "_"],
      FontWeight->Bold],
     TraditionalForm], ")"}], "+", 
   RowBox[{"2", "(", 
    FormBox[
     SuperscriptBox[
      StyleBox[
       OverscriptBox["280", "_"],
       FontWeight->Bold], "\[Prime]"],
     TraditionalForm], ")"}], "+", 
   FormBox[
    StyleBox[
     OverscriptBox["420", "_"],
     FontWeight->Bold],
    TraditionalForm], "+", 
   FormBox[
    SuperscriptBox[
     StyleBox[
      OverscriptBox["450", "_"],
      FontWeight->Bold], "\[Prime]"],
    TraditionalForm], "+", 
   RowBox[{"3", "(", 
    FormBox[
     StyleBox[
      OverscriptBox["480", "_"],
      FontWeight->Bold],
     TraditionalForm], ")"}], "+", 
   RowBox[{"2", "(", 
    FormBox[
     StyleBox[
      OverscriptBox["720", "_"],
      FontWeight->Bold],
     TraditionalForm], ")"}], "+", 
   FormBox[
    StyleBox[
     OverscriptBox["1120", "_"],
     FontWeight->Bold],
    TraditionalForm], "+", 
   FormBox[
    StyleBox[
     OverscriptBox["2520", "_"],
     FontWeight->Bold],
    TraditionalForm]}], TraditionalForm]], "Output",
 ImageSize->{431, 35},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[21]=",
 CellID->542926040]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Decomposition to Subalgebras", "Section",
 CellID->1642176862],

Cell[BoxData[GridBox[{
   {
    RowBox[{
     ButtonBox["DecomposeIrrep",
      BaseStyle->"Link",
      ButtonData->"paclet:LieART/ref/DecomposeIrrep"], "[", 
     RowBox[{
      StyleBox["irrep", "TI"], ",", 
      StyleBox["subalgebra", "TI"]}], "]"}], Cell[TextData[{
     "Decomposes ",
     StyleBox["irrep", "TI"],
     " to the specified ",
     StyleBox["subalgebra.", "TI"]
    }], "TableText"]},
   {
    RowBox[{
     ButtonBox["DecomposeIrrep",
      BaseStyle->"Link",
      ButtonData->"paclet:LieART/ref/DecomposeIrrep"], "[", 
     RowBox[{
      StyleBox["pIrrep", "TI"], ",", 
      StyleBox["subalgebra", "TI"], ",", " ", 
      StyleBox["pos", "TI"]}], "]"}], Cell[TextData[{
     "Decompose ",
     StyleBox["pirrep", "TI"],
     " at position ",
     StyleBox["pos", "TI"],
     " of the product irrep ",
     StyleBox["pIrrep.", "TI"]
    }], "TableText"]}
  }]], "DefinitionBox",
 CellID->1279752208],

Cell["Decompose irreps and product irreps.", "Caption",
 CellID->1925211260],

Cell[TextData[{
 "Decompose the ",
 Cell[BoxData[
  OverscriptBox["10", "_"]], "InlineFormula",
  FontWeight->"Bold"],
 " of SU(5) to SU(3)\[CircleTimes]SU(2)\[CircleTimes]U(1):"
}], "MathCaption",
 CellID->442942768],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DecomposeIrrep", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"Irrep", "[", "SU5", "]"}], "[", 
    RowBox[{"Bar", "[", "10", "]"}], "]"}], ",", 
   RowBox[{"ProductAlgebra", "[", 
    RowBox[{"SU3", ",", "SU2", ",", "U1"}], "]"}]}], "]"}]], "Input",
 CellLabel->"In[22]:=",
 CellID->1304419061],

Cell[BoxData[
 FormBox[
  RowBox[{
   RowBox[{"(", 
    FormBox[
     StyleBox["1",
      FontWeight->Bold],
     TraditionalForm], ",", 
    FormBox[
     StyleBox["1",
      FontWeight->Bold],
     TraditionalForm], ")", "\[InvisibleSpace]", 
    FormBox["\<\"(6)\"\>",
     TraditionalForm]}], "+", 
   RowBox[{"(", 
    FormBox[
     StyleBox["3",
      FontWeight->Bold],
     TraditionalForm], ",", 
    FormBox[
     StyleBox["1",
      FontWeight->Bold],
     TraditionalForm], ")", "\[InvisibleSpace]", 
    FormBox["\<\"(-4)\"\>",
     TraditionalForm]}], "+", 
   RowBox[{"(", 
    FormBox[
     StyleBox[
      OverscriptBox["3", "_"],
      FontWeight->Bold],
     TraditionalForm], ",", 
    FormBox[
     StyleBox["2",
      FontWeight->Bold],
     TraditionalForm], ")", "\[InvisibleSpace]", 
    FormBox["\<\"(1)\"\>",
     TraditionalForm]}]}], TraditionalForm]], "Output",
 ImageSize->{195, 16},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[22]=",
 CellID->682204732]
}, Open  ]],

Cell[TextData[{
 "Decompose the ",
 Cell[BoxData[
  StyleBox["10",
   FontWeight->"Bold"]], "InlineFormula"],
 " and the ",
 Cell[BoxData[
  StyleBox[
   OverscriptBox["5", "_"],
   FontWeight->"Bold"]], "InlineFormula"],
 " of SU(5) to SU(3)\[CircleTimes]SU(2)\[CircleTimes]U(1) (",
 Cell[BoxData[
  ButtonBox["DecomposeIrrep",
   BaseStyle->"Link",
   ButtonData->"paclet:LieART/ref/DecomposeIrrep"]], "InlineFormula"],
 " is ",
 Cell[BoxData[
  ButtonBox["Listable",
   BaseStyle->"Link",
   ButtonData->"paclet:ref/Listable"]], "InlineFormula"],
 "):"
}], "MathCaption",
 CellID->455264099],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DecomposeIrrep", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      RowBox[{"Irrep", "[", "SU5", "]"}], "[", "10", "]"}], ",", 
     RowBox[{
      RowBox[{"Irrep", "[", "SU5", "]"}], "[", 
      RowBox[{"Bar", "[", "5", "]"}], "]"}]}], "}"}], ",", 
   RowBox[{"ProductAlgebra", "[", 
    RowBox[{"SU3", ",", "SU2", ",", "U1"}], "]"}]}], "]"}]], "Input",
 CellLabel->"In[23]:=",
 CellID->2123210495],

Cell[BoxData[
 FormBox[
  RowBox[{"{", 
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      FormBox[
       StyleBox[
        OverscriptBox["3", "_"],
        FontWeight->Bold],
       TraditionalForm], ",", 
      FormBox[
       StyleBox["1",
        FontWeight->Bold],
       TraditionalForm], ")", "\[InvisibleSpace]", 
      FormBox["\<\"(4)\"\>",
       TraditionalForm]}], "+", 
     RowBox[{"(", 
      FormBox[
       StyleBox["3",
        FontWeight->Bold],
       TraditionalForm], ",", 
      FormBox[
       StyleBox["2",
        FontWeight->Bold],
       TraditionalForm], ")", "\[InvisibleSpace]", 
      FormBox["\<\"(-1)\"\>",
       TraditionalForm]}], "+", 
     RowBox[{"(", 
      FormBox[
       StyleBox["1",
        FontWeight->Bold],
       TraditionalForm], ",", 
      FormBox[
       StyleBox["1",
        FontWeight->Bold],
       TraditionalForm], ")", "\[InvisibleSpace]", 
      FormBox["\<\"(-6)\"\>",
       TraditionalForm]}]}], ",", 
    RowBox[{
     RowBox[{"(", 
      FormBox[
       StyleBox[
        OverscriptBox["3", "_"],
        FontWeight->Bold],
       TraditionalForm], ",", 
      FormBox[
       StyleBox["1",
        FontWeight->Bold],
       TraditionalForm], ")", "\[InvisibleSpace]", 
      FormBox["\<\"(-2)\"\>",
       TraditionalForm]}], "+", 
     RowBox[{"(", 
      FormBox[
       StyleBox["1",
        FontWeight->Bold],
       TraditionalForm], ",", 
      FormBox[
       StyleBox["2",
        FontWeight->Bold],
       TraditionalForm], ")", "\[InvisibleSpace]", 
      FormBox["\<\"(3)\"\>",
       TraditionalForm]}]}]}], "}"}], TraditionalForm]], "Output",
 ImageSize->{348, 18},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[23]=",
 CellID->1464885262]
}, Open  ]],

Cell[TextData[{
 "Decompose the ",
 StyleBox["16",
  FontWeight->"Bold"],
 " of SO(10) to SU(5)\[CircleTimes]U(1):"
}], "MathCaption",
 CellID->1947924234],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DecomposeIrrep", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"Irrep", "[", "SO10", "]"}], "[", "16", "]"}], ",", 
   RowBox[{"ProductAlgebra", "[", 
    RowBox[{"SU5", ",", "U1"}], "]"}]}], "]"}]], "Input",
 CellLabel->"In[24]:=",
 CellID->277260301],

Cell[BoxData[
 FormBox[
  RowBox[{
   RowBox[{"(", 
    FormBox[
     StyleBox["1",
      FontWeight->Bold],
     TraditionalForm], ")", "\[InvisibleSpace]", 
    FormBox["\<\"(-5)\"\>",
     TraditionalForm]}], "+", 
   RowBox[{"(", 
    FormBox[
     StyleBox[
      OverscriptBox["5", "_"],
      FontWeight->Bold],
     TraditionalForm], ")", "\[InvisibleSpace]", 
    FormBox["\<\"(3)\"\>",
     TraditionalForm]}], "+", 
   RowBox[{"(", 
    FormBox[
     StyleBox["10",
      FontWeight->Bold],
     TraditionalForm], ")", "\[InvisibleSpace]", 
    FormBox["\<\"(-1)\"\>",
     TraditionalForm]}]}], TraditionalForm]], "Output",
 ImageSize->{162, 16},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[24]=",
 CellID->1904598081]
}, Open  ]],

Cell[TextData[{
 "Decompose the ",
 StyleBox["27",
  FontWeight->"Bold"],
 " of ",
 Cell[BoxData[
  SubscriptBox["E", "6"]], "InlineFormula"],
 " to SU(3)\[CircleTimes]SU(3)\[CircleTimes]SU(3):"
}], "MathCaption",
 CellID->906860298],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DecomposeIrrep", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"Irrep", "[", "E6", "]"}], "[", "27", "]"}], ",", 
   RowBox[{"ProductAlgebra", "[", 
    RowBox[{"SU3", ",", "SU3", ",", "SU3"}], "]"}]}], "]"}]], "Input",
 CellLabel->"In[25]:=",
 CellID->471735093],

Cell[BoxData[
 FormBox[
  RowBox[{
   RowBox[{"(", 
    FormBox[
     StyleBox["3",
      FontWeight->Bold],
     TraditionalForm], ",", 
    FormBox[
     StyleBox["1",
      FontWeight->Bold],
     TraditionalForm], ",", 
    FormBox[
     StyleBox["3",
      FontWeight->Bold],
     TraditionalForm], ")"}], "+", 
   RowBox[{"(", 
    FormBox[
     StyleBox["1",
      FontWeight->Bold],
     TraditionalForm], ",", 
    FormBox[
     StyleBox["3",
      FontWeight->Bold],
     TraditionalForm], ",", 
    FormBox[
     StyleBox[
      OverscriptBox["3", "_"],
      FontWeight->Bold],
     TraditionalForm], ")"}], "+", 
   RowBox[{"(", 
    FormBox[
     StyleBox[
      OverscriptBox["3", "_"],
      FontWeight->Bold],
     TraditionalForm], ",", 
    FormBox[
     StyleBox[
      OverscriptBox["3", "_"],
      FontWeight->Bold],
     TraditionalForm], ",", 
    FormBox[
     StyleBox["1",
      FontWeight->Bold],
     TraditionalForm], ")"}]}], TraditionalForm]], "Output",
 ImageSize->{186, 16},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[25]=",
 CellID->1147868677]
}, Open  ]],

Cell[TextData[{
 "Decompose the SU(3) irrep ",
 StyleBox["3",
  FontWeight->"Bold"],
 " in (",
 StyleBox["24",
  FontWeight->"Bold"],
 ",",
 StyleBox["3",
  FontWeight->"Bold"],
 ")(-3) of SU(5)\[CircleTimes]SU(3)\[CircleTimes]U(1) to \
SU(2)\[CircleTimes]U'(1), i.e., SU(5)\[CircleTimes]SU(3)\[CircleTimes]U(1)\
\[Rule]SU(5)\[CircleTimes]SU(2)\[CircleTimes]U'(1)\[CircleTimes]U(1):"
}], "MathCaption",
 CellID->1532918075],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DecomposeIrrep", "[", 
  RowBox[{
   RowBox[{"ProductIrrep", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"Irrep", "[", "SU5", "]"}], "[", "24", "]"}], ",", 
     RowBox[{
      RowBox[{"Irrep", "[", "SU3", "]"}], "[", "3", "]"}], ",", 
     RowBox[{
      RowBox[{"Irrep", "[", "U1", "]"}], "[", 
      RowBox[{"-", "3"}], "]"}]}], "]"}], ",", 
   RowBox[{"ProductAlgebra", "[", 
    RowBox[{"SU2", ",", "U1"}], "]"}], ",", "2"}], "]"}]], "Input",
 CellLabel->"In[26]:=",
 CellID->1650450684],

Cell[BoxData[
 FormBox[
  RowBox[{
   RowBox[{"(", 
    FormBox[
     StyleBox["24",
      FontWeight->Bold],
     TraditionalForm], ",", 
    FormBox[
     StyleBox["1",
      FontWeight->Bold],
     TraditionalForm], ")", "\[InvisibleSpace]", 
    FormBox["\<\"(-2)\"\>",
     TraditionalForm], "\[InvisibleSpace]", 
    FormBox["\<\"(-3)\"\>",
     TraditionalForm]}], "+", 
   RowBox[{"(", 
    FormBox[
     StyleBox["24",
      FontWeight->Bold],
     TraditionalForm], ",", 
    FormBox[
     StyleBox["2",
      FontWeight->Bold],
     TraditionalForm], ")", "\[InvisibleSpace]", 
    FormBox["\<\"(1)\"\>",
     TraditionalForm], "\[InvisibleSpace]", 
    FormBox["\<\"(-3)\"\>",
     TraditionalForm]}]}], TraditionalForm]], "Output",
 ImageSize->{200, 16},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[26]=",
 CellID->195059642]
}, Open  ]],

Cell["The same decomposition as above displayed as branching rule:", \
"MathCaption",
 CellID->981618749],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"IrrepRule", "[", 
    RowBox[{"#", ",", 
     RowBox[{"DecomposeIrrep", "[", 
      RowBox[{"#", ",", 
       RowBox[{"ProductAlgebra", "[", 
        RowBox[{"SU2", ",", "U1"}], "]"}], ",", "2"}], "]"}]}], "]"}], "&"}], 
  "@", 
  RowBox[{"ProductIrrep", "[", 
   RowBox[{
    RowBox[{
     RowBox[{"Irrep", "[", "SU5", "]"}], "[", "24", "]"}], ",", 
    RowBox[{
     RowBox[{"Irrep", "[", "SU3", "]"}], "[", "3", "]"}], ",", 
    RowBox[{
     RowBox[{"Irrep", "[", "U1", "]"}], "[", 
     RowBox[{"-", "3"}], "]"}]}], "]"}]}]], "Input",
 CellLabel->"In[27]:=",
 CellID->1394022133],

Cell[BoxData[
 FormBox[
  RowBox[{
   RowBox[{"(", 
    FormBox[
     StyleBox["24",
      FontWeight->Bold],
     TraditionalForm], ",", 
    FormBox[
     StyleBox["3",
      FontWeight->Bold],
     TraditionalForm], ")", "\[InvisibleSpace]", 
    FormBox["\<\"(-3)\"\>",
     TraditionalForm]}], "\[RightArrow]", 
   RowBox[{
    RowBox[{"(", 
     FormBox[
      StyleBox["24",
       FontWeight->Bold],
      TraditionalForm], ",", 
     FormBox[
      StyleBox["1",
       FontWeight->Bold],
      TraditionalForm], ")", "\[InvisibleSpace]", 
     FormBox["\<\"(-2)\"\>",
      TraditionalForm], "\[InvisibleSpace]", 
     FormBox["\<\"(-3)\"\>",
      TraditionalForm]}], "+", 
    RowBox[{"(", 
     FormBox[
      StyleBox["24",
       FontWeight->Bold],
      TraditionalForm], ",", 
     FormBox[
      StyleBox["2",
       FontWeight->Bold],
      TraditionalForm], ")", "\[InvisibleSpace]", 
     FormBox["\<\"(1)\"\>",
      TraditionalForm], "\[InvisibleSpace]", 
     FormBox["\<\"(-3)\"\>",
      TraditionalForm]}]}]}], TraditionalForm]], "Output",
 ImageSize->{287, 16},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[27]=",
 CellID->74008952]
}, Open  ]],

Cell["\<\
Branching rules for all totally antisymmetric irreps, so-called basic irreps, \
of SU(6) to SU(3)\[CircleTimes]SU(3)\[CircleTimes]U(1):\
\>", "MathCaption",
 CellID->1987934892],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"IrrepRule", "[", 
     RowBox[{"#", ",", 
      RowBox[{"DecomposeIrrep", "[", 
       RowBox[{"#", ",", 
        RowBox[{"ProductAlgebra", "[", 
         RowBox[{"SU3", ",", "SU3", ",", "U1"}], "]"}]}], "]"}]}], "]"}], 
    "&"}], "/@", 
   RowBox[{"BasicIrreps", "[", "SU6", "]"}]}], "//", "TableForm"}]], "Input",
 CellLabel->"In[28]:=",
 CellID->1514165693],

Cell[BoxData[
 GraphicsBox[
  TagBox[RasterBox[CompressedData["
1:eJztnWlIFV8fx13KLmWmRotZRoVB+eALQSLaRQnbnj+9qTfRk0a3FSEoTYIK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    "], {{0, 96}, {300, 0}}, {0, 255},
    ColorFunction->RGBColor],
   BoxForm`ImageTag["Byte", ColorSpace -> "RGB", Interleaving -> True],
   Selectable->False],
  BaseStyle->"ImageGraphics",
  ImageSize->Automatic,
  ImageSizeRaw->{300, 96},
  PlotRange->{{0, 300}, {0, 96}}]], "Output",
 ImageSize->{304, 96},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[28]//TableForm=",
 CellID->1323284212]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Young Tableaux", "Section",
 CellID->431069882],

Cell[TextData[{
 "The irreps of SU(N) have a correspondence to Young tableaux, which can be \
displayed by ",
 Cell[BoxData[
  ButtonBox["YoungTableau",
   BaseStyle->"Link",
   ButtonData->"paclet:LieART/ref/YoungTableau"]], "InlineFormula"],
 "."
}], "Text",
 CellID->1099666059],

Cell[BoxData[GridBox[{
   {
    RowBox[{
     ButtonBox["YoungTableau",
      BaseStyle->"Link",
      ButtonData->"paclet:LieART/ref/YoungTableau"], "[", 
     StyleBox["irrep", "TI"], "]"}], Cell[TextData[{
     "Displays the Young tableau associated with ",
     StyleBox["irrep", "TI"],
     " "
    }], "TableText"]}
  }]], "DefinitionBox",
 CellID->1528399319],

Cell["Young tableaux.", "Caption",
 CellID->1616075793],

Cell[TextData[{
 "Young tableau of the ",
 StyleBox["720",
  FontWeight->"Bold"],
 " of SU(5):"
}], "MathCaption",
 CellID->381847914],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"YoungTableau", "[", 
  RowBox[{
   RowBox[{"Irrep", "[", "A", "]"}], "[", 
   RowBox[{"1", ",", "2", ",", "0", ",", "1"}], "]"}], "]"}]], "Input",
 CellLabel->"In[29]:=",
 CellID->1550153586],

Cell[BoxData[
 GraphicsBox[
  TagBox[RasterBox[CompressedData["
1:eJzt2MEJwlAUBdGAldiQC+3ABux/FWMLhplH+HMh4ObwmLjL/f15vm/btr2O
53E8v9/7MttW2pm3dCFYKX3Xh5XSd31YKX3Xh5XSd31YKX3Xh5XSd31YKX3X
h5XSd31YKX3Xh0uVrrP/3tLJNyzDKVspB6dspRycspVycMpWysEpWykHp2yl
HJyylXJwylbKwSlbKQf3C35KOlM6Yv1VSlt/ldLWX6W09Vcpbf1VSlt/ldLW
X6W09Vcpbf1VSlt/U191/C31n45Yf5XS1l+ltPVXKW39VUpbf5XS1l+ltPVX
KW39VUpbf+t8XVlnX7zgrWw=
    "], {{0, 83}, {78, 0}}, {0, 255},
    ColorFunction->RGBColor],
   BoxForm`ImageTag["Byte", ColorSpace -> "RGB", Interleaving -> True],
   Selectable->False],
  BaseStyle->"ImageGraphics",
  ImageSize->Automatic,
  ImageSizeRaw->{78, 83},
  PlotRange->{{0, 78}, {0, 83}}]], "Output",
 ImageSize->{82, 83},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[29]=",
 CellID->815317016]
}, Open  ]],

Cell["\<\
Display Young tableaux of SU(4) irreps with a maximum of one column per box \
count:\
\>", "MathCaption",
 CellID->1293870584],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Row", "[", 
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{"Row", "[", 
      RowBox[{"{", 
       RowBox[{"#", ",", "\"\<: \>\"", ",", 
        RowBox[{"YoungTableau", "[", "#", "]"}]}], "}"}], "]"}], "&"}], "/@", 
    
    RowBox[{"SortBy", "[", 
     RowBox[{
      RowBox[{
       RowBox[{"Irrep", "[", "A", "]"}], "@@@", 
       RowBox[{"Tuples", "[", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"0", ",", "1"}], "}"}], ",", "3"}], "]"}]}], ",", "Dim"}], 
     "]"}]}], ",", 
   RowBox[{"Spacer", "[", "10", "]"}]}], "]"}]], "Input",
 CellLabel->"In[30]:=",
 CellID->794726788],

Cell[BoxData[
 GraphicsBox[
  TagBox[RasterBox[CompressedData["
1:eJztm0toVFkQhjsZk4hGRR18EDVMJAER3Ehwo8xGFPEFs9GdEIKKINkJ7rIQ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    "], {{0, 63}, {532, 0}}, {0, 255},
    ColorFunction->RGBColor],
   BoxForm`ImageTag["Byte", ColorSpace -> "RGB", Interleaving -> True],
   Selectable->False],
  BaseStyle->"ImageGraphics",
  ImageSize->Automatic,
  ImageSizeRaw->{532, 63},
  PlotRange->{{0, 532}, {0, 63}}]], "Output",
 ImageSize->{536, 63},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[30]=",
 CellID->441100734]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[" ", "FooterCell"]
},
Saveable->False,
ScreenStyleEnvironment->"Working",
WindowSize->{725, 750},
WindowMargins->{{0, Automatic}, {Automatic, 0}},
WindowTitle->"LieART - Quick Start Tutorial - Wolfram Mathematica",
TaggingRules->{
 "ModificationHighlight" -> False, 
  "Metadata" -> {
   "built" -> "{2012, 7, 17, 8, 55, 57.583545}", "context" -> "LieART`", 
    "keywords" -> {}, "index" -> True, "label" -> 
    "Lie A R T Package Tutorial", "language" -> "en", "paclet" -> 
    "LieART Package", "status" -> "None", "summary" -> 
    "The application LieART provides functions to perform the most common \
tasks associated with irreducible representations (irreps) of the classical \
and exceptional Lie algebas. This loads the package: Irreducible \
representations (irreps) are internally decribed by their Dynkin label with a \
combined head of Irrep and the Lie algebra. ", "synonyms" -> {}, "title" -> 
    "LieART - Quick Start Tutorial", "type" -> "Tutorial", "uri" -> 
    "LieART/tutorial/QuickStartTutorial"}, "LinkTrails" -> "", 
  "SearchTextTranslated" -> ""},
FrontEndVersion->"8.0 for Linux x86 (64-bit) (October 10, 2011)",
StyleDefinitions->Notebook[{
   Cell[
    StyleData[
    StyleDefinitions -> FrontEnd`FileName[{"Wolfram"}, "Reference.nb"]]], 
   Cell[
    StyleData["Input"], CellContext -> "Global`"], 
   Cell[
    StyleData["Output"], CellContext -> "Global`"], 
   Cell[
    StyleData["TutorialMoreAboutSection"], 
    CellGroupingRules -> {"SectionGrouping", 30}], 
   Cell[
    StyleData["RelatedTutorialsSection"], 
    CellGroupingRules -> {"SectionGrouping", 30}], 
   Cell[
    StyleData["TutorialRelatedLinksSection"], 
    CellGroupingRules -> {"SectionGrouping", 30}]}, Visible -> False, 
  FrontEndVersion -> "8.0 for Linux x86 (64-bit) (October 10, 2011)", 
  StyleDefinitions -> "Default.nb"]
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{}
*)
(*CellTagsIndex
CellTagsIndex->{}
*)
(*NotebookFileOutline
Notebook[{
Cell[622, 21, 78, 1, 3, "TutorialColorBar"],
Cell[703, 24, 213, 5, 19, "LinkTrail"],
Cell[919, 31, 2752, 61, 45, "AnchorBarGrid",
 CellID->1],
Cell[CellGroupData[{
Cell[3696, 96, 66, 1, 53, "Title",
 CellID->509267359],
Cell[3765, 99, 222, 5, 54, "Text",
 CellID->1534169418],
Cell[3990, 106, 67, 1, 32, "MathCaption",
 CellID->1167244701],
Cell[CellGroupData[{
Cell[4082, 111, 95, 3, 26, "Input",
 CellID->1092165306],
Cell[CellGroupData[{
Cell[4202, 118, 139, 3, 22, "Print",
 CellID->1715216683],
Cell[4344, 123, 150, 3, 24, "Print",
 CellID->1183076603]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[4543, 132, 76, 1, 55, "Section",
 CellID->1449799695],
Cell[4622, 135, 221, 6, 54, "Text",
 CellID->374349434],
Cell[4846, 143, 448, 15, 59, "DefinitionBox",
 CellID->2096742444],
Cell[5297, 160, 72, 1, 32, "Caption",
 CellID->1891092685],
Cell[5372, 163, 1413, 53, 76, "Text",
 CellID->144687180],
Cell[6788, 218, 223, 8, 32, "MathCaption",
 CellID->836781195],
Cell[CellGroupData[{
Cell[7036, 230, 209, 7, 28, "Input",
 CellID->2058623809],
Cell[7248, 239, 435, 15, 37, "Output",
 CellID->710845293]
}, Open  ]],
Cell[7698, 257, 261, 8, 32, "MathCaption",
 CellID->1221898746],
Cell[CellGroupData[{
Cell[7984, 269, 212, 7, 28, "Input",
 CellID->711449157],
Cell[8199, 278, 185, 5, 50, "Output",
 CellID->320085412]
}, Open  ]],
Cell[8399, 286, 257, 8, 32, "MathCaption",
 CellID->681119358],
Cell[CellGroupData[{
Cell[8681, 298, 172, 5, 28, "Input",
 CellID->937078051],
Cell[8856, 305, 245, 9, 35, "Output",
 CellID->829231914]
}, Open  ]],
Cell[9116, 317, 574, 16, 76, "Text",
 CellID->319667636],
Cell[9693, 335, 287, 11, 32, "MathCaption",
 CellID->1881934180],
Cell[CellGroupData[{
Cell[10005, 350, 196, 6, 28, "Input",
 CellID->464169775],
Cell[10204, 358, 222, 8, 35, "Output",
 CellID->1931618218]
}, Open  ]],
Cell[10441, 369, 352, 7, 76, "Text",
 CellGroupingRules->{"GroupTogetherGrouping", 31},
 CellID->1468912962],
Cell[10796, 378, 433, 15, 59, "DefinitionBox",
 CellID->792167334],
Cell[11232, 395, 76, 1, 32, "Caption",
 CellID->1579371899],
Cell[11311, 398, 351, 13, 32, "MathCaption",
 CellID->109632626],
Cell[CellGroupData[{
Cell[11687, 415, 195, 6, 28, "Input",
 CellID->333587606],
Cell[11885, 423, 181, 5, 48, "Output",
 CellID->1988021587]
}, Open  ]],
Cell[12081, 431, 105, 2, 32, "MathCaption",
 CellID->1511260992],
Cell[CellGroupData[{
Cell[12211, 437, 197, 6, 28, "Input",
 CellID->1425383264],
Cell[12411, 445, 180, 5, 48, "Output",
 CellID->622885791]
}, Open  ]],
Cell[12606, 453, 347, 9, 54, "Text",
 CellID->76966508],
Cell[12956, 464, 478, 15, 59, "DefinitionBox",
 CellID->616137037],
Cell[13437, 481, 54, 1, 32, "Caption",
 CellID->722483301],
Cell[13494, 484, 276, 11, 32, "MathCaption",
 CellID->1501046753],
Cell[CellGroupData[{
Cell[13795, 499, 294, 9, 28, "Input",
 CellID->1686390499],
Cell[14092, 510, 390, 16, 37, "Output",
 CellID->1102757659]
}, Open  ]],
Cell[CellGroupData[{
Cell[14519, 531, 101, 3, 28, "Input",
 CellID->680021955],
Cell[14623, 536, 205, 5, 48, "Output",
 CellID->1927209351]
}, Open  ]],
Cell[CellGroupData[{
Cell[14865, 546, 105, 2, 26, "Input",
 CellID->1308426208],
Cell[14973, 550, 389, 16, 37, "Output",
 CellID->739296930]
}, Open  ]],
Cell[15377, 569, 226, 5, 47, "MathCaption",
 CellID->1021666716],
Cell[CellGroupData[{
Cell[15628, 578, 362, 11, 28, "Input",
 CellID->1109740049],
Cell[15993, 591, 437, 17, 37, "Output",
 CellID->1746525800]
}, Open  ]],
Cell[CellGroupData[{
Cell[16467, 613, 103, 3, 28, "Input",
 CellID->1050432190],
Cell[16573, 618, 228, 8, 63, "Output",
 CellID->1160495567]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[16850, 632, 119, 2, 55, "Section",
 CellGroupingRules->{"SectionGrouping", Inherited},
 CellID->354599395],
Cell[16972, 636, 365, 11, 59, "DefinitionBox",
 CellID->1309055104],
Cell[17340, 649, 69, 1, 32, "Caption",
 CellID->1795200765],
Cell[17412, 652, 278, 11, 32, "MathCaption",
 CellID->1189993846],
Cell[CellGroupData[{
Cell[17715, 667, 299, 9, 28, "Input",
 CellID->1332771530],
Cell[18017, 678, 355, 15, 35, "Output",
 CellID->1163229237]
}, Open  ]],
Cell[18387, 696, 339, 14, 32, "MathCaption",
 CellID->133991959],
Cell[CellGroupData[{
Cell[18751, 714, 303, 9, 28, "Input",
 CellID->101963549],
Cell[19057, 725, 439, 19, 35, "Output",
 CellID->600695207]
}, Open  ]],
Cell[19511, 747, 314, 12, 32, "MathCaption",
 CellID->1984894909],
Cell[CellGroupData[{
Cell[19850, 763, 342, 10, 28, "Input",
 CellID->1356021807],
Cell[20195, 775, 470, 20, 37, "Output",
 CellID->502898059]
}, Open  ]],
Cell[20680, 798, 253, 10, 32, "MathCaption",
 CellID->1005626329],
Cell[CellGroupData[{
Cell[20958, 812, 267, 8, 28, "Input",
 CellID->1887184167],
Cell[21228, 822, 664, 29, 37, "Output",
 CellID->1871092227]
}, Open  ]],
Cell[21907, 854, 654, 23, 47, "MathCaption",
 CellID->683723966],
Cell[CellGroupData[{
Cell[22586, 881, 102, 3, 28, "Input",
 CellID->997413459],
Cell[22691, 886, 278, 9, 78, "Output",
 CellID->107401848]
}, Open  ]],
Cell[22984, 898, 543, 18, 47, "MathCaption",
 CellID->1581190875],
Cell[CellGroupData[{
Cell[23552, 920, 102, 3, 28, "Input",
 CellID->863826808],
Cell[23657, 925, 762, 33, 39, "Output",
 CellID->1351507635]
}, Open  ]],
Cell[24434, 961, 87, 1, 32, "MathCaption",
 CellID->328377256],
Cell[CellGroupData[{
Cell[24546, 966, 771, 23, 62, "Input",
 CellID->451431300],
Cell[25320, 991, 1319, 59, 37, "Output",
 CellID->814463409]
}, Open  ]],
Cell[26654, 1053, 490, 19, 32, "MathCaption",
 CellID->1480557077],
Cell[CellGroupData[{
Cell[27169, 1076, 579, 18, 45, "Input",
 CellID->343055154],
Cell[27751, 1096, 582, 25, 37, "Output",
 CellID->415443930]
}, Open  ]],
Cell[28348, 1124, 375, 14, 32, "MathCaption",
 CellID->846894030],
Cell[CellGroupData[{
Cell[28748, 1142, 433, 12, 28, "Input",
 CellID->1274938883],
Cell[29184, 1156, 1234, 52, 39, "Output",
 CellID->1329380349]
}, Open  ]],
Cell[30433, 1211, 341, 13, 32, "MathCaption",
 CellID->163980459],
Cell[CellGroupData[{
Cell[30799, 1228, 375, 11, 28, "Input",
 CellID->62108204],
Cell[31177, 1241, 2151, 89, 56, "Output",
 CellID->542926040]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[33377, 1336, 68, 1, 55, "Section",
 CellID->1642176862],
Cell[33448, 1339, 925, 31, 81, "DefinitionBox",
 CellID->1279752208],
Cell[34376, 1372, 76, 1, 32, "Caption",
 CellID->1925211260],
Cell[34455, 1375, 217, 7, 32, "MathCaption",
 CellID->442942768],
Cell[CellGroupData[{
Cell[34697, 1386, 315, 9, 28, "Input",
 CellID->1304419061],
Cell[35015, 1397, 1023, 41, 37, "Output",
 CellID->682204732]
}, Open  ]],
Cell[36053, 1441, 594, 22, 32, "MathCaption",
 CellID->455264099],
Cell[CellGroupData[{
Cell[36672, 1467, 439, 13, 45, "Input",
 CellID->2123210495],
Cell[37114, 1482, 1757, 67, 39, "Output",
 CellID->1464885262]
}, Open  ]],
Cell[38886, 1552, 155, 6, 32, "MathCaption",
 CellID->1947924234],
Cell[CellGroupData[{
Cell[39066, 1562, 271, 8, 28, "Input",
 CellID->277260301],
Cell[39340, 1572, 768, 29, 37, "Output",
 CellID->1904598081]
}, Open  ]],
Cell[40123, 1604, 233, 9, 32, "MathCaption",
 CellID->906860298],
Cell[CellGroupData[{
Cell[40381, 1617, 282, 8, 28, "Input",
 CellID->471735093],
Cell[40666, 1627, 1119, 49, 37, "Output",
 CellID->1147868677]
}, Open  ]],
Cell[41800, 1679, 423, 14, 47, "MathCaption",
 CellID->1532918075],
Cell[CellGroupData[{
Cell[42248, 1697, 519, 15, 45, "Input",
 CellID->1650450684],
Cell[42770, 1714, 876, 33, 37, "Output",
 CellID->195059642]
}, Open  ]],
Cell[43661, 1750, 105, 2, 32, "MathCaption",
 CellID->981618749],
Cell[CellGroupData[{
Cell[43791, 1756, 631, 20, 45, "Input",
 CellID->1394022133],
Cell[44425, 1778, 1197, 45, 37, "Output",
 CellID->74008952]
}, Open  ]],
Cell[45637, 1826, 187, 4, 32, "MathCaption",
 CellID->1987934892],
Cell[CellGroupData[{
Cell[45849, 1834, 421, 13, 45, "Input",
 CellID->1514165693],
Cell[46273, 1849, 5080, 91, 130, "Output",
 CellID->1323284212]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[51402, 1946, 53, 1, 55, "Section",
 CellID->431069882],
Cell[51458, 1949, 281, 9, 54, "Text",
 CellID->1099666059],
Cell[51742, 1960, 366, 12, 59, "DefinitionBox",
 CellID->1528399319],
Cell[52111, 1974, 55, 1, 32, "Caption",
 CellID->1616075793],
Cell[52169, 1977, 134, 6, 32, "MathCaption",
 CellID->381847914],
Cell[CellGroupData[{
Cell[52328, 1987, 215, 6, 28, "Input",
 CellID->1550153586],
Cell[52546, 1995, 814, 21, 104, "Output",
 CellID->815317016]
}, Open  ]],
Cell[53375, 2019, 136, 4, 32, "MathCaption",
 CellID->1293870584],
Cell[CellGroupData[{
Cell[53536, 2027, 623, 21, 45, "Input",
 CellID->794726788],
Cell[54162, 2050, 3323, 62, 82, "Output",
 CellID->441100734]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[57524, 2117, 23, 0, 42, "FooterCell"]
}
]
*)

(* End of internal cache information *)

